/* Author: kurt.stephens@parlano.com 2000/10/06 */
/*   parlano eventloop library -- pl_el/eventloop.h
 *   Copyright (C) 2001 Parlano, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _pl_el_h
#define _pl_el_h

#include <sys/time.h>

#include "pl_fd/pl_fd.h"

#include "tevent.h"
#include "callback.h"

struct pl_el_delayed_free_entry;
typedef struct pl_el_delayed_free_entry pl_el_delayed_free_entry;

struct pl_el_delayed_free_entry {
  pl_el_delayed_free_entry *next;
  /*@ Next link in delayed_free_chain. */

  void *ptr;
  /*@ The data to free(). */
};

struct pl_el {
  /*@ The event loop object. */

  pl_fd **fdobjs;
  /*@ A list of registered fd objs. */

  int nfdobjs;
  /*@ Number of registered fd objs. */

  int running;
  /*@ The event loop is running. */

  int stop;
  /*@ Stops the event loop. */

  int stop_poll;
  /*@ Stops the current event poll. */

  int result;
  /*@ The result for pl_el_run(). */

  struct timeval timeout[2];
  /*@ The timeout for select(). */

  struct timeval *timeoutp;
  /*@ The timeoutp for select(). */

  /* Callbacks */

  pl_cb begin_run;
  /*@ Called at begining of pl_el_run() */

  pl_cb begin_poll;
  /*@ Called at beginning of pl_el_poll() */

  pl_cb before_select;
  /*@ Called before select(). */

  pl_cb after_select;
  /*@ Called after select(). */

  pl_cb end_poll;
  /*@ Called at end of pl_el_poll() */

  pl_cb end_run;
  /*@ Called at end of pl_el_run() */

  pl_el_delayed_free_entry *delayed_free_list;
  /*@ Chain of things to free() at end of pl_el_poll(). */

  pl_tevent_q teq;
  /*@ Timed event queue. */

  int debug;
  /*@ Debug support. */

  void *opaque[2];
  /*@ Opaque data. */
};


/*****************************************************/
/* Init, destroy. */

int pl_el_init(pl_el *el);
int pl_el_destroy(pl_el *el);
int pl_el_deepen(pl_el *el);

/*****************************************************/
/* FD set manip. */

int pl_el_add(pl_el *el, pl_fd *fdo);
int pl_el_remove(pl_el *el, pl_fd *fdo);

/*****************************************************/
/* Event loop execution. */

pl_el *pl_el_current();
int pl_el_poll(pl_el *el);
int pl_el_poll_stop(pl_el *el);
int pl_el_run(pl_el *el);
pl_el *pl_el_running();
int pl_el_stop(pl_el *el, int result);

/*****************************************************/
/* Callbacks. */

#define pl_el_callback_handler_DECL(X) int X (pl_el *el, void *data, void *handle)
typedef pl_el_callback_handler_DECL((*pl_el_callback_handler));

void *pl_el_begin_run(pl_el *el, pl_el_callback_handler func, void *data, int repeat);
void *pl_el_begin_poll(pl_el *el, pl_el_callback_handler func, void *data, int repeat);
void *pl_el_before_select(pl_el *el, pl_el_callback_handler func, void *data, int repeat);
void *pl_el_after_select(pl_el *el, pl_el_callback_handler func, void *data, int repeat);
void *pl_el_end_poll(pl_el *el, pl_el_callback_handler func, void *data, int repeat);
void *pl_el_end_run(pl_el *el, pl_el_callback_handler func, void *data, int repeat);

int pl_el_callback_remove(pl_el *el, void *handle);

void *pl_el_at_time(pl_el *el, pl_tevent_time time, pl_el_callback_handler func, void *data);

int pl_el_at_time_remove(pl_el *el, void *handle);

/*****************************************************/
/* delayed free. */

void pl_el_delayed_free(pl_el *el, void *ptr);
void pl_delayed_free(void *ptr);

/*****************************************************/
/* EOF */


#endif
