/* Author: kurt.stephens@parlano.com 2000/10/03 */
/*   parlano eventloop library -- pl_fd/fdobj.h
 *   Copyright (C) 2001 Parlano, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _pl_fd_h
#define _pl_fd_h

#ifndef pl_osfd
/* Assume POSIX "int" file-descriptors. */
#define pl_osfd int
#define pl_osfd_INVALID -1
#define pl_osfd_IS_VALID(X) ((X) >= 0)
#endif

#include <stddef.h> /* size_t */
#include <sys/socket.h>
#include <netinet/in.h> /* struct sockaddr_in */

struct pl_fd;
typedef struct pl_fd pl_fd;

struct pl_el;
typedef struct pl_el pl_el;

#define pl_fd_event_handler_DECL(X) int X (pl_fd *fdo, pl_el *el)
typedef pl_fd_event_handler_DECL((*pl_fd_event_handler));

typedef enum pl_fd_error {
  /*@ pl_fd_*() return error codes. */

  pl_fd_ERROR = -1,
  /*@ An OS-level error occurred, see fdo->oserr. */

  pl_fd_EWOULDBLOCK = -2,
  /*@ The OS-level system call would have blocked. */

  pl_fd_EINPROGRESS = -3,
  /*@ The OS-level system call would have blocked, but is now in progress. */

  pl_fd_EALREADY = -4
  /*@ The OS-level system call is already in progress. */
} pl_fd_error;


enum pl_fd_event {
  /*@ select() event mask. */

  pl_fd_EXCEPT = 1,     /*@ %pl_fd% event_except handler. */
  pl_fd_WRITE = 2,      /*@ %pl_fd% event_write handler. */
  pl_fd_READ = 4,       /*@ %pl_fd% event_read handler. */

  pl_fd_OSEXCEPT = 8,   /*@ %pl_fd% event_osexcept handler. */
  pl_fd_OSWRITE = 16,   /*@ %pl_fd% event_oswrite handler. */
  pl_fd_OSREAD = 32     /*@ %pl_fd% event_osread handler. */
};


typedef enum pl_fd_state {
  /*@ pl_fd states. */

  pl_fd_DESTROYED = -1,
  /*@ The object was pl_fd_destroy()ed. */

  pl_fd_INVALID = 0,
  /*@ OS-level fd is invalid or not created. */

  pl_fd_CLOSED,
  /*@ OS-level fd was closed. */

  pl_fd_HAS_FD,
  /*@ OS-level fd has been created. */
  
  pl_fd_OPENED,
  /*@ OS-level socket has been opened. */

  pl_fd_BOUND,
  /*@ OS-level socket has been bound. */

  pl_fd_LISTENING,
  /* OS-level socket is listening. */

  pl_fd_CONNECTING,
  /*@ OS-level socket is connecting. */

  pl_fd_CONNECTED,
  /*@ OS-level socket is connected. */

  pl_fd_ACCEPTING,
  /*@ OS-level bound socket is accepting. */

  pl_fd_ACCEPTED
  /*@ OS-level bound socket has accepted. */

} pl_fd_state;


struct pl_fd {
  /*@ A file descriptor abstraction object. */

  pl_osfd fd;     
  /*@ The OS-level file handle. */

  pl_fd_state state;
  /*@ The current state of the file handle. */

  int oserr;
  /*@ The last OS-level error on this object. */

  const char *oserrstr;
  /*@ The last OS-level error on this object in human readable format. */

  int (*osread)(pl_fd *fdo, void *buf, size_t size);
  /*@ The OS-level read method. */

  int (*oswrite)(pl_fd *fdo, const void *buf, size_t size);
  /*@ The OS-level write method. */
  
  int (*osopen)(pl_fd *fd);
  /*@ The OS-level open method. */

  int (*osclose)(pl_fd *fd);
  /*@ The OS-level close method. */

  int (*osshutdown)(pl_fd *fd);
  /*@ The OS-level shutdown method. */

  /*********************************************************************/
  /* open parameters. */

  char *name;
  /*@ The name of the fd. */

  char *local_name;
  /*@ The local name of the fd. */

  int nblock;
  /*@ Non-blocking I/O. */

  int osflags;
  /*@ The OS-level fcntl() flags. */

  int osmode;
  /*@ The OS-level mode of the open fd. */

  int oscreat;
  /*@ The OS-level file creation mask. */

  /*********************************************************************/
  /* Socket support. */

  struct sockaddr_in ossockaddr;
  /*@ Socket address. */

  struct sockaddr_in os_local_sockaddr;
  /*@ Local socket address for bind(). */

  struct sockaddr_in os_send_addr;
  /*@ Socket address for sendto(). */

  struct sockaddr_in os_recv_addr;
  /*@ Socket address for recvfrom(). */

  int ossockdomain;
  /*@ Socket family: PF_INET. */

  int ossocktype;
  /*@ Socket type: SOCK_STREAM, SOCK_DGRAM, etc. */

  int osproto;
  /*@ Socket proto: IPPROTO_TCP, IPPROTO_UDP, etc. */

  pl_fd *accept_from;
  /*@ Accepted from... */

  pl_fd *proto;
  /*@ Use proto to initialize asynchonously accepted clients. */

  /*********************************************************************/
  /* eventloop support. */

  int event_enable;
  /*@ Event handlers enabled. */

  int event_mask;
  /*@ Event handler mask for read, write and except. */

  pl_fd_event_handler event_osread;
  /*@ OS-level select() says "read" is avail. */

  pl_fd_event_handler event_oswrite;
  /*@ OS-level select() says "write" is avail. */

  pl_fd_event_handler event_osexcept;
  /*@ OS-level select() says "read" is avail. */
  
  pl_el *el;
  /*@ Current event loop. */

  /*********************************************************************/
  /* State notification. */

  pl_fd_event_handler event_read;
  /*@ Read event handler. */

  pl_fd_event_handler event_write;
  /*@ Write event handler. */

  pl_fd_event_handler event_except;
  /*@ Exception event handler. */

  /*  XXX:  DO NOT USE THIS FOR ANYTHING.  IT IS A HACK
   *        WITH VERY FEW CHECKS AROUND IT.  YOU HAVE 
   *        BEEN WARNED IN ALL CAPS.
   */
  pl_fd_event_handler __save_event_except;
  
  pl_fd_event_handler event_accepting;
  /*@ Called after moved to ACCEPTING state. */
  
  pl_fd_event_handler event_accepted;
  /*@ Called after moved to ACCEPTED state. */

  pl_fd_event_handler event_accept_error;
  /*@ Called after a accept connection error. */

  pl_fd_event_handler event_connecting;
  /*@ Called after moved to CONNECTING state. */

  pl_fd_event_handler event_connected;
  /*@ Called after moved to CONNECTED state. */

  pl_fd_event_handler event_connect_error;
  /*@ Called after a connection error. */

  pl_fd_event_handler event_opened;
  /*@ Called when pl_fd_opened is sucessful. */

  pl_fd_event_handler event_closing;
  /*@ Called before moved to CLOSED state. */

  pl_fd_event_handler event_retry_open;
  /*@ Called before open retry. */

  pl_fd_event_handler event_retry_open_error;
  /*@ Called if open retry failed. */

  /*********************************************************************/
  /* Asyncronous open retry support. */

  int async;
  /*@ Asynchronus open? */

  int retry_open_max;
  /*@ Maximum number of open retries.  If < 0 keep retrying, if > 0 retrys N times. */

  int retry_open_delay;
  /*@ Seconds to wait before retrying open(). */

  int retry_open_count;
  /*@ Current number of failed open retrys. */

  void *retry_open_timer;
  /*@ Handle for current open retry timer. */

  /*********************************************************************/
  /* user-level support. */

  int (*destroy)(pl_fd *fdo);
  /*@ The user-level destroy method. */

  void (*dealloc)(void *fd);
  /*@ The user-level deallocation method. */

  int debug;
  /*@ Debug support. */

  void *opaque[2]; 
  /*@ Client data. */
};


/********************************************/
/* configuration. */

const char* pl_fd_local_name_default();
int pl_fd_set_local_name_default(const char *addr);

const char *pl_fd_local_name_default_envvar();
int pl_fd_set_local_name_default_envvar(const char *addr);

/********************************************/
/* init and destroy. */

int pl_fd_init(pl_fd *fdo);
int pl_fd_destroy(pl_fd *fdo);

/********************************************/
/* Local address. */

const char *pl_fd_local_name(pl_fd *fdo);
int pl_fd_set_local_name(pl_fd *fdo, const char *addr);


/********************************************/
/* High-level I/O. */

int pl_fd_read(pl_fd *fdo, void *buf, size_t size);
int pl_fd_write(pl_fd *fdo, const void *buf, size_t size);
int pl_fd_open(pl_fd *fdo);
int pl_fd_close(pl_fd *fdo);
int pl_fd_shutdown(pl_fd *fdo);

int pl_fd_reopen(pl_fd *fdo);

/********************************************/
/* High-level options */

int pl_fd_nblock(pl_fd *fdo);
int pl_fd_set_nblock(pl_fd *fdo, int nblock);

/********************************************/
/* State. */

pl_fd_state pl_fd_get_state(pl_fd *fdo);
int pl_fd_get_oserr(pl_fd *fdo);
const char *pl_fd_get_oserrstr(pl_fd *fdo);


/********************************************/
/* Create fdobj of a given type. */

int pl_fd_fd(pl_fd *fdo, pl_osfd fd);

int pl_fd_file(pl_fd *fdo, const char *file, int mode, int creatmask);

int pl_fd_socket(pl_fd *fdo, const char *conn);
int pl_fd_socket_async(pl_fd *fdo, const char *conn);

int pl_fd_socket_server(pl_fd *fdo, const char *conn);
int pl_fd_socket_server_async(pl_fd *fdo, const char *conn, pl_fd *proto);

int pl_fd_accept(pl_fd *fdo, pl_fd *fdos);
int pl_fd_accept_async(pl_fd *fdo, pl_fd *fdos);

/********************************************/
/* Convience. */

char *pl_fd_read_until(pl_fd *fdo, int term);

/********************************************/
/* EOF */

#endif
