/*   parlano eventloop library -- pl_rr/reqres.h
 *   Copyright (C) 2001 Parlano, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _pl_rr_h
#define _pl_rr_h

/* Author: kurt.stephens@parlano.com 2000/09/26 */

/*************************************************************************/

#include <stddef.h> /* size_t */
#include <stdio.h> /* FILE */

typedef unsigned long pl_rr_id;

/*************************************************************************/
/* Forward decls. */

struct pl_rr;
typedef struct pl_rr pl_rr;

struct pl_rr_endpoint;
typedef struct pl_rr_endpoint pl_rr_endpoint;

struct pl_rr_q;
typedef struct pl_rr_q pl_rr_q;

#define pl_rr_handler_DECL(X) int X (pl_rr *rr, void *data, pl_rr_endpoint *ep)
typedef pl_rr_handler_DECL((*pl_rr_handler));

/*************************************************************************/

#include "buf.h"
#include "endpoint.h"

/*************************************************************************/

typedef enum pl_rr_direction {
  /*@ What direction is the request? */

  pl_rr_OUTGOING = 0,
  /*@ Requests going out. */

  pl_rr_INCOMING = 1,
  /*@ Requests coming in. */

} pl_rr_direction;


#define pl_rr_DST pl_rr_OUTGOING /*@ Destination object for the request. */
#define pl_rr_SRC pl_rr_INCOMING /*@ Source object for the request. */


typedef enum pl_rr_op {
  /*@ The request/response duality. */

  pl_rr_REQUEST = pl_rr_OUTGOING,
  /*@ This is a request. */

  pl_rr_RESPONSE = pl_rr_INCOMING
  /*@ This is a response. */
} pl_rr_op;


typedef enum pl_rr_state {
  pl_rr_DESTROYED = -1,
  /*@ The object has been destroyed. */

  pl_rr_START = 0,
  /*@ The object has been initialized. */

  pl_rr_REQUEST_COMPLETE,
  /*@ The request data is complete. */

  pl_rr_REQUEST_IO,
  /*@ The request is being sent. */

  pl_rr_REQUEST_IO_COMPLETE,
  /*@ The request is being sent. */

  pl_rr_RESPONSE_COMPLETE,
  /*@ The response data is complete. */

  pl_rr_RESPONSE_IO,
  /*@ The response is complete. */

  pl_rr_RESPONSE_IO_COMPLETE,
  /*@ The response is being sent. */

  pl_rr_CLOSED,
  /*@ The request/response has been completed. */

  pl_rr_ABORTED,
  /*@ The request/response has been aborted. */

} pl_rr_state;


#define pl_rr_event_DECL(X) int X(pl_rr *rr)
typedef pl_rr_event_DECL((*pl_rr_event_handler));

struct pl_rr {
  /*@ A request/response queue entry. */

  pl_rr_state state;
  /*@ Current state. */

  unsigned long _refcount;
  /*@ Reference count. */

  struct pl_rr *next, *prev;
  /*@ The pl_rr_q linked list ptrs. */

  pl_rr_q *queue;
  /*@ The queue this req/res is in. */

  pl_rr_endpoint *ep[2];
  /*@ Destination[0] or source[1] endpoint of the request. */

  pl_rr_direction dir;
  /*@ The direction of the request. */

  short proto;
  /*@ The reqres protocol version. */

  pl_rr_id id;
  /*@ The request id. */

  pl_rr_buf buf[2];
  /*@ The request/response buffers */

  pl_rr_handler resh;
  /*@ The routine to callback when a response is ready. */

  void *resh_data;
  /*@ The data argument for %resh%. */

  /*******************************************/
  /* Event handlers. */
  
  pl_rr_event_handler event_destroy;
  /*@ 
    Called when the rr is being destroyed. 
    state will be pl_rr_CLOSED or pl_rr_ABORT.
  */

  void *opaque[2];
  /*@ Opaque data. */
};


/*************************************************************************/

#include "queue.h"

/*************************************************************************/

pl_rr_id pl_rr_get_id(pl_rr *rr);
#define pl_rr_get_id(rr) ((rr)->id)

#if 0
pl_rr_buf *pl_rr_req(pl_rr *rr);
#define pl_rr_req(rr) (&(rr)->buf[pl_rr_REQUEST])

pl_rr_buf *pl_rr_res(pl_rr *rr);
#define pl_rr_res(rr) (&(rr)->buf[pl_rr_RESPONSE])
#endif

/*************************************************************************/
/* Request. */

size_t pl_rr_req_len(pl_rr *rr);
const char *pl_rr_req_buf(pl_rr *rr);
pl_rr_id  pl_rr_req_part(pl_rr *rr);
int pl_rr_req_write(pl_rr *rr, const void *buf, size_t len);
int pl_rr_req_close(pl_rr *rr);

/*************************************************************************/
/* Response. */

pl_rr_id pl_rr_res_part(pl_rr *rr);
const char *pl_rr_res_buf(pl_rr *rr);
size_t pl_rr_res_len(pl_rr *rr);
int pl_rr_res_write(pl_rr *rr, const void *buf, size_t len);
int pl_rr_res_read(pl_rr *rr, void *buf, size_t len);
int pl_rr_res_close(pl_rr *rr);

/*************************************************************************/
/* State. */

pl_rr_state pl_rr_get_state(pl_rr *rr);

/*************************************************************************/
/* Endpoints. */

pl_rr_endpoint *pl_rr_dst(pl_rr *rr);
pl_rr_endpoint *pl_rr_src(pl_rr *rr);

/*************************************************************************/
/* Debugging. */

int pl_rr_dump(pl_rr *rr, FILE *fp);

/*************************************************************************/
/* Internals. */

int pl_rr_init(pl_rr *rr);
int pl_rr_req_init(pl_rr *rr, pl_rr_id id, pl_rr_id part, size_t len);
int _pl_rr_destroy(pl_rr *rr);
int _pl_rr_close(pl_rr *rr, int op);


/*************************************************************************/

#endif
