/* Author: kurt.stephens@parlano.com 2000/10/02 */
/*   parlano eventloop library -- pl_el/eventloop.c
 *   Copyright (C) 2001 Parlano, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "eventloop.h"
#include "timeval.h"

#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#ifndef __CYGWIN32__
#include <alloca.h>
#endif

#ifndef PL_EL_DEBUG
#if 0
#define PL_EL_DEBUG 1
#else
#define PL_EL_DEBUG el->debug
#endif
#endif

/*******************************************************************/
/* Hooks for pl_el and pl_tevent_queue */

static int pl_el_before_select_tevent_q(pl_el *el)
     /*@
       Called by event loop event to set up select() timeout.
     */
{
  pl_tevent_q *q = &el->teq;
  assert(q);

  /*
     el->timeout[1] contains the fire time for the next timed event.
     select() needs total time to wait.
     Thus, compute the difference between the time to fire and now.
  */
  /* Compute difference from now and timeout[1] */
  gettimeofday(&el->timeout[0], 0);
  el->timeout[0] = pl_timeval_sub(el->timeout[1], el->timeout[0]);

  return 0;
}



static int pl_el_after_select_tevent_q(pl_el *el)
     /*@
       Called by event loop event to handle timed events.
     */
{
  pl_tevent_q *q = &el->teq;
  pl_tevent_time now;

  assert(el);
  assert(q);

  /* Get the time of now. */
  now = pl_tevent_time_now();

  /* Check for any events to fire. */
  pl_tevent_q_check(q, now);

  return 0;
}


static int pl_tevent_q_eventlist_set_alarm(pl_tevent_q *q, pl_tevent_time when, void (*func)())
     /*@
       Called by timed event queue after modifications.
      */
{
  pl_el *el;

  assert(q);
  el = q->data1;
  assert(el);

  /* when == 0 if nothing is in the queue. */
  if ( pl_tevent_time_to_time_t(when) ) {
    /* Save when in timeout[1]. */
    el->timeout[1] = pl_tevent_time_to_timeval(when);

    /* Prepare select() timeout. */
    el->timeoutp = &el->timeout[0];
  } else {
    /* No select() timeout. */
    el->timeoutp = 0;
  }

  return 0;
}


/*******************************************************************/
/* Delayed free support. */


static int _pl_el_delayed_free_free(pl_el *el)
{
  pl_el_delayed_free_entry *x, *x_next;

  x = el->delayed_free_list;
  el->delayed_free_list = 0;
  for ( ; x; x = x_next ) {
    x_next = x->next;
    if ( PL_EL_DEBUG ) {
      fprintf(stderr, "delayed_free(%p)\n", x->ptr);
    }
    free(x->ptr);
    free(x);
  }

  return 0;
}


void pl_el_delayed_free(pl_el *el, void *ptr)
     /*@ Puts %ptr% on a list to free()ed at the end of %pl_el_poll%() or %pl_el_destroy%(). */
{
  assert(el);
  if ( ptr ) {
    pl_el_delayed_free_entry *x;

    if ( PL_EL_DEBUG ) {
      fprintf(stderr, "pl_el_delayed_free(%p, %p)\n", el, ptr);
    }

    /* Make sure it hasn't been delayed_free() before. */
    if ( PL_EL_DEBUG ) {
      fprintf(stderr, "free_list: ");
    }
    for ( x = el->delayed_free_list; x; x = x->next ) {
      if ( PL_EL_DEBUG ) {
	fprintf(stderr, "%p ", x);
      }
      if ( x->ptr == ptr ) {
	fprintf(stderr, "\npl_el_delayed_free(%p, %p): %p freed more than once\n", el, ptr, ptr);
	assert(x->ptr != ptr);
	abort();
      }
    }
    if ( PL_EL_DEBUG ) {
      fprintf(stderr, "\n");
    }

    /* Add to the eventloop's delayed_free_list. */
    /* Gotta malloc() to free()!!!.  Gotta spend money to make money. */
    x = malloc(sizeof(*x));
    x->next = el->delayed_free_list;
    x->ptr = ptr;
    el->delayed_free_list = x;
  }
}


void pl_delayed_free(void *ptr)
     /*@
       Calls %pl_el_delayed_free%(%pl_el_running%(), %ptr%);
       If no event loop is running, it frees it immediately.
      */
{
  pl_el *el = pl_el_running();

  if ( el ) {
    pl_el_delayed_free(el, ptr);
  } else {
    fprintf(stderr, "pl_delayed_free(%p): WARNING: no eventloop is running\n", ptr);
    free(ptr);
  }
}


/*******************************************************************/


static int _pl_el_init(pl_el *el)
{
  assert(el);

  /* Initialize callback lists. */
  pl_cb_init(&el->begin_run);
  pl_cb_init(&el->begin_poll);
  pl_cb_init(&el->before_select);
  pl_cb_init(&el->after_select);
  pl_cb_init(&el->end_poll);
  pl_cb_init(&el->end_run);

  /* Initialize timed event queue. */
  pl_tevent_q_init(&el->teq);
  el->teq.data1 = el;
  el->teq.set_alarm = pl_tevent_q_eventlist_set_alarm;

  return 0;
}


int pl_el_init(pl_el *el)
     /*@ Initialize the %el%. */
{
  assert(el);

  memset(el, 0, sizeof(*el));

  el->fdobjs = malloc(0);
  el->nfdobjs = 0;

  _pl_el_init(el);

  return 0;
}


int pl_el_deepen(pl_el *el)
     /*@
       Gives %el% new copy of its fdobjs.
      All other lists are reinitialized.
     */
{
  void *p;

  assert(el);
  p = malloc(sizeof(el->fdobjs[0]) * el->nfdobjs);
  assert(p);
  memcpy(p, el->fdobjs, sizeof(el->fdobjs[0]) * el->nfdobjs);

  el->fdobjs = p;

  /* It's not running, yet. */
  el->running = 0;

  /* WARNING: DOES NOT DEEPEN CALLBACKS!! */
  /* WARNING: DOES NOT DEEPEN TIMED EVENT QUEUE!!! */
  _pl_el_init(el);

  return 0;
}



int pl_el_destroy(pl_el *el)
     /*@ Destroys %el%.  Does not destroy any %fdobjs%. */
{
  int i;

  assert(el);

  /* If this event loop is running, pl_el_destroy was called too early. */
  assert(el->running == 0);

  /* Remove any associations between any fdo's and el. */
  for ( i = 0; i < el->nfdobjs; ++ i ) {
    pl_fd *fdo = el->fdobjs[i];

    if ( fdo->el == el ) {
      fdo->el = 0;
    }
  }

  free(el->fdobjs);
  el->fdobjs = 0;
  el->nfdobjs = 0;

  pl_cb_destroy(&el->begin_run);
  pl_cb_destroy(&el->begin_poll);
  pl_cb_destroy(&el->before_select);
  pl_cb_destroy(&el->after_select);
  pl_cb_destroy(&el->end_poll);
  pl_cb_destroy(&el->end_run);

  pl_tevent_q_destroy(&el->teq);

  _pl_el_delayed_free_free(el);

  return 0;
}


int pl_el_add(pl_el *el, pl_fd *fdo)
     /*@
       Add a file descriptor to the event list.
       Returns non-zero if %fdo% is already in %el%.
      */
{
  int i;

  assert(el);
  assert(fdo);

  /*@ Check to see if it is already added. */
  for ( i = 0; i < el->nfdobjs; ++ i ) {
    if ( el->fdobjs[i] == fdo ) {
      fdo->el = el;
      return 1;
    }
  }

  /* Look for empty slot. */
  for ( i = 0; i < el->nfdobjs; ++ i ) {
    if ( el->fdobjs[i] == 0 )
      break;
  }

  /* Resize fdobjs. */
  if ( i >= el->nfdobjs ) {
    el->fdobjs = realloc(el->fdobjs, sizeof(el->fdobjs[0]) * (i + 1));
    el->nfdobjs = i + 1;
  }

  el->fdobjs[i] = fdo;
  fdo->el = el;

  return 0;
}


int pl_el_contains(pl_el *el, pl_fd *fdo)
     /*@
       Returns non-zero if %fdo% is in %el%.
      */
{
  int i;

  assert(el);
  assert(fdo);

  for ( i = 0; i < el->nfdobjs; ++ i ) {
    if ( el->fdobjs[i] == fdo ) {
      return 1;
    }
  }

  return 0;
}


int pl_el_remove(pl_el *el, pl_fd *fdo)
     /*@
       Remove a file descriptor from the event list.
       Returns non-zero if %fdo% is not in %el%.
      */
{
  int i;

  assert(el);
  assert(fdo);

  for ( i = 0; i < el->nfdobjs; ++ i ) {
    if ( el->fdobjs[i] == fdo ) {
      /* Remove association between fdo and el */
      if ( fdo->el == el ) {
	fdo->el = 0;
      }

      /* Remove fdobjs entry by sliding down remainder of fdobjs. */
      ++ i;
      for ( ; i < el->nfdobjs; ++ i ) {
        el->fdobjs[i - 1] = el->fdobjs[i];
      }
      el->nfdobjs --;

      /* Terminate any _pl_el_poll that is running. */
      pl_el_poll_stop(el);

      return 0;
    }
  }

  return -1;
}


/****************************************************************************/
/* Callbacks. */

void *pl_el_begin_run(pl_el *el, pl_el_callback_handler func, void *data, int repeat)
{
  assert(el);

  return pl_cb_add(&el->begin_run, (pl_cb_handler) func, data, repeat ? pl_cb_REPEAT : pl_cb_ONESHOT);
}

void *pl_el_begin_poll(pl_el *el, pl_el_callback_handler func, void *data, int repeat)
{
  assert(el);

  return pl_cb_add(&el->begin_poll, (pl_cb_handler) func, data, repeat ? pl_cb_REPEAT : pl_cb_ONESHOT);
}


void *pl_el_before_select(pl_el *el, pl_el_callback_handler func, void *data, int repeat)
{
  assert(el);

  return pl_cb_add(&el->before_select, (pl_cb_handler) func, data, repeat ? pl_cb_REPEAT : pl_cb_ONESHOT);
}


void *pl_el_after_select(pl_el *el, pl_el_callback_handler func, void *data, int repeat)
{
  assert(el);

  return pl_cb_add(&el->after_select, (pl_cb_handler) func, data, repeat ? pl_cb_REPEAT : pl_cb_ONESHOT);
}


void *pl_el_end_poll(pl_el *el, pl_el_callback_handler func, void *data, int repeat)
{
  assert(el);

  return pl_cb_add(&el->end_poll, (pl_cb_handler) func, data, repeat ? pl_cb_REPEAT : pl_cb_ONESHOT);
}


void *pl_el_end_run(pl_el *el, pl_el_callback_handler func, void *data, int repeat)
{
  assert(el);

  return pl_cb_add(&el->end_run, (pl_cb_handler) func, data, repeat ? pl_cb_REPEAT : pl_cb_ONESHOT);
}


int pl_el_callback_remove(pl_el *el, void *handle)
{
  assert(el);

  return pl_cb_remove(handle);
}


/* Timed events. */

void *pl_el_at_time(pl_el *el, pl_tevent_time when, pl_el_callback_handler func, void *data)
{
  assert(el);

  return pl_tevent_q_add(&el->teq, when, (pl_tevent_proc) func, data);
}


int pl_el_at_time_remove(pl_el *el, void *handle)
{
  assert(el);

  if ( handle ) {
    return pl_tevent_q_remove(&el->teq, handle);
  }
  return -1;
}


/****************************************************************************/


#define fd_IS_VALID(fd) (pl_osfd_IS_VALID(fd) && (fd) <= FD_SETSIZE)

static pl_el *el_running = 0; /* NOT THREAD-SAFE!!!. */

pl_el *pl_el_running()
     /*@ The currently running eventloop. */
{
  return el_running;
}


int pl_el_poll_stop(pl_el *el)
     /*@ Stops the current running pl_el_poll() call. */
{
  assert(el);

  el->stop_poll = 1;

  return 0;
}


static int _pl_el_poll(pl_el *el, struct timeval *timeoutp)
     /*@
       Polls for events once.
       If %timeoutp% is null, just poll for events.
      */
{
  int result = 0;
  int nfds = 0;
  int returned_nfds = 0;
  int i;
  pl_el *el_running_save;

  fd_set rfds;
  /*@ The read fd_set. */

  fd_set wfds;
  /*@ The write fd_set. */

  fd_set efds;
  /*@ The exception fd_set. */

  assert(el);

  /* Remember what el was running, remember what el is running now. */
  el_running_save = el_running;
  el_running = el;

  /* Clear stop_poll. */
  el->stop_poll = 0;

  do {
#define CHECK_STOP() \
  if ( el->stop_poll || el->stop ) break

    /* Do begin_poll callbacks. */
    pl_cb_all(&el->begin_poll, el);

    CHECK_STOP();

    /* Clear fd_sets. */
    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    FD_ZERO(&efds);

    /* Put all file descriptors into fd_sets. */
    for ( i = 0; i < el->nfdobjs; ++ i ) {
      pl_fd *fdo;
      int em = 0;

      CHECK_STOP();

      fdo = el->fdobjs[i];

      /* Invalid fdo? */
      if ( fdo == 0 )
	continue;

      /* Invalid fd? */
      if ( ! fd_IS_VALID(fdo->fd) )
	continue;

      /* Events enabled? */
      if ( fdo->event_enable <= 0 )
	continue;

      /* Read events? */
      if ( ((fdo->event_mask & pl_fd_READ) && fdo->event_read) ||
	   ((fdo->event_mask & pl_fd_OSREAD) && fdo->event_osread)
	   ) {
	em |= pl_fd_OSREAD | pl_fd_READ;
	FD_SET(fdo->fd, &rfds);
	if ( nfds < fdo->fd )
	  nfds = fdo->fd;
      }

      /* Write events? */
      if ( ((fdo->event_mask & pl_fd_WRITE) && fdo->event_write) ||
	   ((fdo->event_mask & pl_fd_OSWRITE) && fdo->event_oswrite)
	   ) {
	em |= pl_fd_OSWRITE | pl_fd_WRITE;
	FD_SET(fdo->fd, &wfds);
	if ( nfds < fdo->fd )
	  nfds = fdo->fd;
      }

      /* Exception events? */
      if ( ((fdo->event_mask & pl_fd_EXCEPT) && fdo->event_except) ||
	   ((fdo->event_mask & pl_fd_OSEXCEPT) && fdo->event_osexcept)
	   ) {
	em |= pl_fd_OSEXCEPT | pl_fd_EXCEPT;
	FD_SET(fdo->fd, &efds);
	if ( nfds < fdo->fd )
	  nfds = fdo->fd;
      }

      if ( fdo->debug ) {
	fprintf(stderr, "  FD_SET %-20s SEL %d %c%c%c EN %c OSM %c%c%c OSE %c%c%c M %c%c%c E %c%c%c\n",
		fdo->name,
		fdo->fd,
		em & pl_fd_OSREAD ? 'r' : '-',
		em & pl_fd_OSWRITE ? 'w' : '-',
		em & pl_fd_OSEXCEPT ? 'x' : '-',
		fdo->event_enable ? '+' : '-',
		fdo->event_mask & pl_fd_OSREAD ? 'r' : '-',
		fdo->event_mask & pl_fd_OSWRITE ? 'w' : '-',
		fdo->event_mask & pl_fd_OSEXCEPT ? 'x' : '-',
		fdo->event_osread ? 'r' : '-',
		fdo->event_oswrite ? 'w' : '-',
		fdo->event_osexcept ? 'x' : '-',
		fdo->event_mask & pl_fd_READ ? 'r' : '-',
		fdo->event_mask & pl_fd_WRITE ? 'w' : '-',
		fdo->event_mask & pl_fd_EXCEPT ? 'x' : '-',
		fdo->event_read ? 'r' : '-',
		fdo->event_write ? 'w' : '-',
		fdo->event_except ? 'x' : '-'
		);
      }
    }

    /* Do before_select callbacks. */
    pl_cb_all(&el->before_select, el);

    CHECK_STOP();

   /* Do select() */
    do {
      /* Handle timed events. */
      pl_el_before_select_tevent_q(el);

      CHECK_STOP();

      if ( PL_EL_DEBUG ) {
	fprintf(stderr, "%p: select(%d, ...)\n", el, nfds + 1);
      }

      if ( PL_EL_DEBUG && timeoutp ) {
	fprintf(stderr, "%p: timeout %ld.%06lu\n", el, (long) timeoutp->tv_sec, (unsigned long) timeoutp->tv_usec);
      }

      result = returned_nfds = select(nfds + 1, &rfds, &wfds, &efds, timeoutp);
    } while ( result < 0 && errno == EINTR );

    if ( PL_EL_DEBUG && result < 0 ) {
      fprintf(stderr, "select(): returned %d: %d: %s\n", result, errno, strerror(errno));
    }

    /* Handle timed events. */
    pl_el_after_select_tevent_q(el);

    CHECK_STOP();

    /* Do after_select callbacks. */
    pl_cb_all(&el->after_select, el);

    CHECK_STOP();

    if ( result > 0 ) {
      /* Copy el->fdobjs to avoid event handler side-effects on el. */
      int nfdobjs = el->nfdobjs;
      pl_fd **fdobjs = alloca(sizeof(fdobjs[0]) * nfdobjs);

      memcpy(fdobjs, el->fdobjs, sizeof(fdobjs[0]) * nfdobjs);

      result = 0;

      /* For each fdobj. */
      for ( i = 0; i < nfdobjs; ++ i ) {
	pl_fd *fdo = fdobjs[i];
	int x;
	int em;
	pl_el *el_save;

	CHECK_STOP();

	if ( fdo == 0 )
	  continue;

	/* fdo may have been removed from el by a previous event. */
	if ( ! pl_el_contains(el, fdo) )
	  continue;

	/* Invalid fd? */
	if ( ! fd_IS_VALID(fdo->fd) )
	  continue;

	/* Save fdo's event loop, incase there are nested event loops. */
	el_save = fdo->el;

	do {
	  /* Compute select() result eventmask. */

	  /* Assume that el's FD_SETS are never side-effected by event handlers. */
	  em =
	    (FD_ISSET(fdo->fd, &rfds) ? (pl_fd_OSREAD | pl_fd_READ) : 0) |
	    (FD_ISSET(fdo->fd, &wfds) ? (pl_fd_OSWRITE | pl_fd_WRITE) : 0) |
	    (FD_ISSET(fdo->fd, &efds) ? (pl_fd_OSEXCEPT | pl_fd_EXCEPT) : 0);

	  if ( ! em )
	    break;

	  if ( PL_EL_DEBUG || fdo->debug ) {
	    fprintf(stderr, "  %-20s SEL %d %c%c%c EN %c OSM %c%c%c OSE %c%c%c M %c%c%c E %c%c%c\n",
		  fdo->name,
		    fdo->fd,
		    em & pl_fd_OSREAD ? 'r' : '-',
		    em & pl_fd_OSWRITE ? 'w' : '-',
		    em & pl_fd_OSEXCEPT ? 'x' : '-',
		    fdo->event_enable ? '+' : '-',
		    fdo->event_mask & pl_fd_OSREAD ? 'r' : '-',
		    fdo->event_mask & pl_fd_OSWRITE ? 'w' : '-',
		    fdo->event_mask & pl_fd_OSEXCEPT ? 'x' : '-',
		    fdo->event_osread ? 'r' : '-',
		    fdo->event_oswrite ? 'w' : '-',
		    fdo->event_osexcept ? 'x' : '-',
		    fdo->event_mask & pl_fd_READ ? 'r' : '-',
		    fdo->event_mask & pl_fd_WRITE ? 'w' : '-',
		    fdo->event_mask & pl_fd_EXCEPT ? 'x' : '-',
		    fdo->event_read ? 'r' : '-',
		    fdo->event_write ? 'w' : '-',
		    fdo->event_except ? 'x' : '-'
		    );
	  }

#define DO_IT(E,F) ((fdo->event_enable > 0) && (fdo->event_mask & E & em) && fdo->F)

	  /*
	    NOTE:
	    OS-level events preempt non-OS-level events, so user doesn't have to
	    know any pecularities of socket behavior during connect() and accept() as
	    seen by select().
	    In other words: all the OS- events should never be seen by application
	    code.
	  */

	  /* Error. */
	  if ( DO_IT(pl_fd_OSEXCEPT,event_osexcept) ) {
	    if ( PL_EL_DEBUG ) {
	      fprintf(stderr, "    event_osexcept(%s, %p)\n", fdo->name, el);
	    }
	    x = (fdo->event_osexcept)(fdo, el);
	    if ( x ) {
	      break;
	    }
	  } else
	    if ( DO_IT(pl_fd_EXCEPT,event_except) ) {
	      if ( PL_EL_DEBUG ) {
		fprintf(stderr, "    event_except(%s, %p)\n", fdo->name, el);
	      }
	      x = (fdo->event_except)(fdo, el);
	      if ( x ) {
		break;
	      }
	    }

	  CHECK_STOP();


	  /* Read. */
	  if ( DO_IT(pl_fd_OSREAD,event_osread) ) {
	    if ( PL_EL_DEBUG ) {
	      fprintf(stderr, "    event_osread(%s, %p)\n", fdo->name, el);
	    }

	    x = (fdo->event_osread)(fdo, el);
	    if ( x ) {
	      break;
	    }
	  } else
	    if ( DO_IT(pl_fd_READ,event_read) ) {
	      if ( PL_EL_DEBUG ) {
		fprintf(stderr, "    event_read(%s, %p)\n", fdo->name, el);
	      }

	      x = (fdo->event_read)(fdo, el);
	      if ( x ) {
		break;
	      }
	    }

	  CHECK_STOP();

	  /* Write. */
	  if ( DO_IT(pl_fd_OSWRITE,event_oswrite) ) {
	    if ( PL_EL_DEBUG ) {
	      fprintf(stderr, "    event_oswrite(%s, %p)\n", fdo->name, el);
	    }
	    x = (fdo->event_oswrite)(fdo, el);
	    if ( x ) {
	      break;
	    }
	  } else
	    if ( DO_IT(pl_fd_WRITE,event_write) ) {
	      if ( PL_EL_DEBUG ) {
		fprintf(stderr, "    event_write(%s, %p)\n", fdo->name, el);
	      }
	      x = (fdo->event_write)(fdo, el);
	      if ( x ) {
		break;
	      }
	    }

	} while ( 0 );

#undef DO_IT

	/* Restore fdo's event loop. */
	fdo->el = el_save;

	CHECK_STOP();
      }
    }

    CHECK_STOP();

    /* Do end_poll callbacks(). */
    pl_cb_all(&el->end_poll, el);

#undef CHECK_STOP

  } while(0);

  /* Not running anymore. */
  el_running = el_running_save;

  /* Do delayed free list. */
  _pl_el_delayed_free_free(el);

  return result;
}


int pl_el_poll(pl_el *el)
     /*@
       Poll for any events and return immediatly.
     */
{
  struct timeval tv = { 0, 0 };

  return _pl_el_poll(el, &tv);
}


int pl_el_run(pl_el *el)
     /*@
       Runs event loop until %pl_el_stop%.
      */
{
  int result;

  assert(el);

  if ( el->running ) {
    return -1;
  }
  ++ el->running;

  el->stop = 0;

  pl_cb_all(&el->begin_run, el);

  do {
    result = _pl_el_poll(el, el->timeoutp);
    /* Continue. */
  } while ( result == 0 && ! el->stop );

  -- el->running;

  pl_cb_all(&el->end_run, el);

  return el->result;
}


int pl_el_stop(pl_el *el, int result)
     /*@ Stops running event loop and causes %pl_el_run% to return with %result%. */
{
  assert(el);
  assert(el->running);

  el->stop = 1;
  el->result = result;

  return 0;
}


