/* Author: kurt.stephens@parlano.com 2000/10/06 */
/*   parlano eventloop library -- pl_el/tevent.c
 *   Copyright (C) 2001 Parlano, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "tevent.h"

#include <signal.h> /* signal() */
#include <unistd.h> /* alarm() */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#ifndef PL_TEVENT_FUDGE
#define PL_TEVENT_FUDGE 0
/*@ Add PL_TEVENT_FUDGE to the alarm() argument to handle discrepencies. */
#endif

#ifndef PL_TEVENT_USE_ALARM
#define PL_TEVENT_USE_ALARM 1
/*@ If non-zero use alarm()/signal() to schedule. */
#endif


/**************************************************************************
 ** Unit testing.
 */

#ifdef UNIT_TEST

#ifndef PL_TEVENT_DEBUG
#define PL_TEVENT_DEBUG 0
#endif

static void user_alarm2(int sig)
{
  printf("user_alarm2()\n");
}

static void user_alarm(int sig)
{
  printf("user_alarm()\n");
  signal(SIGALRM, user_alarm2);
  alarm(7);
}

static pl_tevent_proc_DECL(proc)
{
  printf("proc('%s')\n", (char *) data2);
}

static pl_tevent_proc_DECL(again)
{
  static int count = 6;
  int dt = (int) data2;
  printf("again(%d)\n", dt);
  if ( -- count == 0 ) {
    printf("DONE.");
    exit(0);
  }

  pl_tevent_q_add(data1, pl_time_t_to_tevent_time(time(0) + dt), again, (void*) dt);
}


int main(int argc, char **argv)
{
  time_t start = time(0);
  pl_tevent_q _q, *q = &_q;
  memset(q, 0, sizeof(*q));

  /* Define user alarm. */
  signal(SIGALRM, user_alarm);
  alarm(8);

  pl_tevent_q_init(q);

  pl_tevent_q_add(q, pl_time_t_to_tevent_time(start + 10), proc, "10");
  pl_tevent_q_add(q, pl_time_t_to_tevent_time(start + 5), proc, "5");
  pl_tevent_q_add(q, pl_time_t_to_tevent_time(start + 12), proc, "12");
  pl_tevent_q_add(q, pl_time_t_to_tevent_time(start + 6), again, (void*) 6);

  while ( 1 ) {
    sleep(2);
    printf("%d secs elapsed\n", (int) (time(0) - start));
  }

  return 0;
}

#endif


/**************************************************************************
 ** Debugging support.
 */

#ifndef PL_TEVENT_DEBUG
#define PL_TEVENT_DEBUG 0
/*@ If non-zero enable debugging output. */
#endif

/**************************************************************************
 ** Queue management.
 */

#define pl_tevent_q_begin(q) ((q)->queue.next)
#define pl_tevent_q_end(q)   (&(q)->queue)


static void pl_tevent_print(pl_tevent_q *q, FILE *fp, pl_tevent *x)
     /*@ Print event entry %x% to %fp%. */
{
  if ( x->prev ) {
    assert(pl_tevent_time_cmp(x->when, x->prev->when) >= 0);
  }
  fprintf(fp,
	  "  <-%p %p ->%p %lu (now+%lu) %p(%p)\n", 
	  (void *) x->prev,
	  (void *) x,
	  (void *) x->next,
	  (unsigned long) pl_tevent_time_to_time_t(x->when), 
	  (unsigned long) (pl_tevent_time_to_time_t(x->when) - time(0)),
	  (void *) x->proc, 
	  (void *) x->data2
	  );
}


static void pl_tevent_q_print(pl_tevent_q *q, FILE *fp)
     /*@ Prints entire queue to %fp%. */
{
  pl_tevent *x;

  fprintf(fp, "tevent at %lu {\n", (unsigned long) time(0));
  for ( x = pl_tevent_q_begin(q); x != pl_tevent_q_end(q); x = x->next ) {
    assert(x->next);
    assert(x->prev);
    pl_tevent_print(q, fp, x);
  }
  fprintf(fp, "}\n");
}


static void pl_tevent_q_remove_1 (pl_tevent_q *q, pl_tevent *l)
     /*@ Removes entry %l% from the queue. */
{
  assert(l);
  if ( l->queue ) {
    assert(l->queue == q);
    l->queue = 0;
  }
  if ( l->next && l->prev ) {
    l->prev->next = l->next;
    l->next->prev = l->prev;
#if PL_TEVENT_DEBUG
    fprintf(stderr, "Removed "); pl_tevent_print(q, stderr, l);
    pl_tevent_q_print(q, stderr);
#endif
  }
  l->next = l->prev = 0;
}


static void pl_tevent_q_add_after (pl_tevent_q *q, pl_tevent *l, pl_tevent *after)
     /*@ Added entry %l% to the queue after %after%. */
{
  assert(q);
  assert(l);
  assert(after);

  /* Remove from list first. */
  pl_tevent_q_remove_1(q, l);

  l->queue = q;
  l->prev = after;
  l->next = after->next;

#if PL_TEVENT_DEBUG
  if ( ! after->next ) {
    fprintf(stderr, "Adding "); pl_tevent_print(q, stderr, l);
    fprintf(stderr, "after->next == 0; after = "); pl_tevent_print(q, stderr, after);
  }
#endif
  assert(after->next);
  assert(after->prev);

  after->next->prev = l;
  after->next = l;

#if PL_TEVENT_DEBUG
  fprintf(stderr, "Added "); pl_tevent_print(q, stderr, l);
#endif
}


static void pl_tevent_q_remove_event_1(pl_tevent_q *q, pl_tevent *l)
     /*@ Removes entry %l% from the queue and destroys it. */
{
  assert(l);

  if ( ! q ) {
    q = l->queue;
  }

  if ( l->queue ) {
    pl_tevent_q_remove_1(q, l);
    if ( l->destroy ) {
#if 0
      void (*d)(void*) = l->destroy;
      memset(l, 0, sizeof(*l));
      (d)(l);
#else
      l->destroy(l);
#endif
    }
  }
}


/************************************************************************
 ** Low-level alarm() support.
 */

static void pl_tevent_q_alarm_handler();

static int pl_tevent_q_set_alarm(pl_tevent_q *q);

static int pl_tevent_q_set_alarm_alarm(pl_tevent_q *q, pl_tevent_time at_time, void (*func)());

PL_TEVENT_EXPORT int (*pl_tevent_q_set_alarm_default)(pl_tevent_q *q, pl_tevent_time at_time, void (*func)()) = &pl_tevent_q_set_alarm_alarm;


/*
** Implements alarm scheduling using POSIX alarm()/signal().  
** Could be recoded using another event loop.
*/

/* Handle external alarm handler chaning. */
static pl_tevent_proc_DECL(pl_tevent_q_signal_user)
{
#if PL_TEVENT_DEBUG
  fprintf(stderr, "pl_tevent_q_signal_user()\n");
#endif

  /* Chain to prev handler. */
#if 0
  signal(SIGALRM, data2);
  raise(SIGALRM);
#else
  if ( ! (data2 == (void*) SIG_IGN || data2 == (void*) SIG_DFL || data2 == (void*) SIG_ERR) ) {
    assert(data2);
    ((void(*)(int)) data2)(SIGALRM);
  }
#endif

  /* Reset our alarm handler. */
  pl_tevent_q_set_alarm(data1);
}



static unsigned pl_alarm(unsigned sec)
     /*@ Calls system alarm(). */
{
  unsigned left = alarm(sec);

#if PL_TEVENT_DEBUG
  fprintf(stderr, "tevent: alarm(%u) = %u\n", sec, left);
#endif

  return left;
}

#define alarm(X) pl_alarm(X)

static int pl_tevent_q_in_handler = 0;
/*@ Non-zero if the handler is being processed. */

static pl_tevent_q *pl_tevent_q_alarm_handler_queue;
/*@ The last %pl_tevent_q% to call %pl_tevent_set_alarm_alarm%. */


static int pl_tevent_q_set_alarm_alarm(pl_tevent_q *q, pl_tevent_time at_time, void (*func)())
     /*@
       Sets alarm() and signal() handler.
       Checks for previous alarm() handler.
      */
{
  static int recursion_lock;
  static pl_tevent_time alarm_at;


#if PL_TEVENT_DEBUG
  fprintf(stderr, "pl_tevent_set_alarm_alarm() %lu.%06lu\n",
	  (unsigned long) pl_tevent_time_to_time_t(at_time),
	  (unsigned long) pl_tevent_time_to_usec(at_time));
#endif

  if ( recursion_lock ) {
#if PL_TEVENT_DEBUG
    fprintf(stderr, "pl_tevent_set_alarm_alarm() recursion lock\n");
#endif
    return 1;
  }

  pl_tevent_q_alarm_handler_queue = q;
  assert(q);

  ++ recursion_lock;
  assert(recursion_lock);

  do { 
    pl_tevent_time now = pl_tevent_time_now();

#if PL_TEVENT_DEBUG
    fprintf(stderr, "pl_tevent_set_alarm_alarm() recursion lock\n");
#endif

#if 0
    /* If the alarm was already set, dont set it again. */
    if ( pl_tevent_time_to_time_t(alarm_at) == pl_tevent_time_to_time_t(at_time) ) {
#if PL_TEVENT_DEBUG
      fprintf(stderr, "pl_tevent_set_alarm_alarm() already set to %lu\n",
	      (unsigned long) (alarm_at));
#endif
      break;
    }
#endif

    alarm_at = at_time;

  /* Set-up an alarm. */
#if PL_TEVENT_USE_ALARM
#ifdef SIGALRM
    assert(func);

    if ( pl_tevent_time_to_time_t(at_time) == 0 ) {
#if PL_TEVENT_DEBUG
      fprintf(stderr, "Ignoring null at_time\n");
#endif
      alarm(0);
    } else {
      void *signal_prev;
      unsigned int alarm_prev;
      
      
      signal_prev = (void*) 
#if 1
	signal
#else
	sigset
#endif
	(SIGALRM, func);

#ifdef SIG_ERR
      assert(signal_prev != (void*) SIG_ERR);
#endif

      if ( pl_tevent_time_cmp(now, at_time) >= 0 ) {
	at_time = pl_time_t_to_tevent_time(1);
      } else {
	at_time = pl_tevent_time_sub(at_time, now);
      }

      {
	time_t x_at_time = pl_tevent_time_to_time_t(at_time);

#ifdef PL_TEVENT_FUDGE
	x_at_time += PL_TEVENT_FUDGE;
#endif

	if ( ! x_at_time )
	  ++ x_at_time;

#if PL_TEVENT_DEBUG
	fprintf(stderr, "alarm(%lu)\n", (unsigned long) x_at_time);
#endif
	/* Begin alarm */
	alarm_prev = pl_alarm(x_at_time);

#ifdef sun
	/*
	** Adjust alarm() such that the expected time matches the requested time.
	** See: solaris_alarm_bug.c for details.
	** Calling alarm() from abritrary signal handlers does not work.
	*/
	{
	  unsigned dt = x_at_time;
	  unsigned expected;
	  
	  while ( (expected = pl_alarm(dt)) != x_at_time ) {
	    if ( expected > x_at_time ) {
	      -- dt;
	    } else {
	      ++ dt;
	    }
	  }
	}
#endif
	
#if 0
	/* Chain prev alarm handler if alarm() was called by someone else. */
	if ( pl_tevent_in_handler == 0 && alarm_prev > 0 && signal_prev != pl_tevent_alarm_handler ) {
	  alarm_prev += pl_tevent_time_to_time_t(now);
#if PL_TEVENT_DEBUG
	  fprintf(stderr, "chaining signal_prev = %p at %lu\n", 
		  (void*) signal_prev,
		  (unsigned long) alarm_prev);
	  fprintf(stderr, "pl_tevent_default_alarm = %p\n", (void*) pl_tevent_alarm_default);
#endif
	  pl_tevent_q_add(q, pl_time_t_to_tevent_time(alarm_prev), (void*) pl_tevent_q_signal_user, (void*) signal_prev);
	}
#endif
      }
    }

#else
#error SIGALRM not defined!
#endif
#endif  
    
  } while ( 0 );

  assert(recursion_lock);
  -- recursion_lock;

  return 0;
}



static int pl_tevent_q_set_alarm(pl_tevent_q *q)
     /*@ Sets alarm handler to trigger on first event in queue. */
{
  pl_tevent *l;

  /* Reset alarm handler for next event. */
  if ( (l = pl_tevent_q_begin(q)) != pl_tevent_q_end(q) ) {
    assert(pl_tevent_time_to_time_t(l->when));
    return q->set_alarm(q, l->when, pl_tevent_q_alarm_handler);
  } else {
    /* Disable the alarm handler. */
    assert(pl_tevent_time_to_time_t(q->queue.when) == 0);
    return q->set_alarm(q, q->queue.when, pl_tevent_q_alarm_handler);
  }
}


int pl_tevent_q_check(pl_tevent_q *q, pl_tevent_time now)
{
  int removed = 0;
  pl_tevent *l;

  /* Scan head of queue for expired events. */
  while ( ((l = pl_tevent_q_begin(q)) != pl_tevent_q_end(q)) && pl_tevent_time_cmp(l->when, now) <= 0 ) {
    assert(pl_tevent_time_to_time_t(l->when) != 0);

    /*
    ** Trigger and remove. 
    */
    (l->proc)(q->data1, l->data2, (void*) l);
    
    pl_tevent_q_remove_1(q, l);

    ++ removed;
  }

  if ( removed ) {
    /* Reset alarm handler. */
    pl_tevent_q_set_alarm(q);
  }

  return 0;
}


static void pl_tevent_q_alarm_handler()
     /*@
       The handler for SIGALRM.
       Process expired events from the head of the queue.
     */
{
  pl_tevent_time now;
  pl_tevent_q *q = pl_tevent_q_alarm_handler_queue;

  assert(q);

  if ( pl_tevent_q_in_handler ) {
#if PL_TEVENT_DEBUG
    fprintf(stderr, "pl_tevent_alarm_handler() called recursively\n");
#endif
    return;
  }

  ++ pl_tevent_q_in_handler;
  assert(pl_tevent_q_in_handler);

  now = pl_tevent_time_now();

  pl_tevent_q_check(q, now);

#if PL_TEVENT_DEBUG
  fprintf(stderr, "pl_tevent_alarm_handler() now = %lu\n", (unsigned long) pl_tevent_time_to_time_t(now));
#endif

#if PL_TEVENT_DEBUG
  fprintf(stderr, "after SIGALARM handler: \n");
  pl_tevent_q_print(q, stderr);
#endif

  assert(pl_tevent_q_in_handler);
  -- pl_tevent_q_in_handler;
}


PL_TEVENT_EXPORT int pl_tevent_q_add_event (pl_tevent_q *q, pl_tevent *l)
     /*@ 
       Adds an event entry to the queue and initializes the alarm handler.
     */
{
  assert(q);
  assert(l);
  assert(pl_tevent_time_to_time_t(l->when));
  assert(l->proc);

  do { 
    pl_tevent *x;

    /* Remove first. */
    pl_tevent_q_remove_1(q, l);

    /* Scan queue for insert point. */
    for ( x = pl_tevent_q_begin(q); x != pl_tevent_q_end(q); x = x->next ) {
      assert(x != l);
      /*
      ** &pl_tevent   
      ** [0] ------- [20] ------- [30] ------  pl_tevent_end()
      ** l ---            [24]
      */
      if ( pl_tevent_time_cmp(l->when, x->when) < 0 )
	break;
    }
      
    /* Add to queue before x. */
    pl_tevent_q_add_after(q, l, x->prev);
    
#if PL_TEVENT_DEBUG
    pl_tevent_q_print(q, stderr);
#endif

    /* Set alarm. */
    pl_tevent_q_set_alarm(q);
  } while ( 0 );

  return 0;
}


/*******************************************************************************/

PL_TEVENT_EXPORT int pl_tevent_q_init(pl_tevent_q *q)
     /*@ Initialize the %pl_tevent_q%. */
{
  assert(q);

  memset(q, 0, sizeof(*q));

  q->queue.next = q->queue.prev = &q->queue;
  if ( ! q->set_alarm ) {
    q->set_alarm = pl_tevent_q_set_alarm_default;
  }

  return 0;
}


PL_TEVENT_EXPORT int pl_tevent_q_destroy(pl_tevent_q *q)
     /*@ Destroy %q%. */
{
  pl_tevent *x;

  assert(q);

  while ( (x = pl_tevent_q_begin(q)) != pl_tevent_q_end(q) ) {
    pl_tevent_q_remove_event_1(q, x);
  }

  return 0;
}


PL_TEVENT_EXPORT void *pl_tevent_q_add(pl_tevent_q *q, pl_tevent_time when, pl_tevent_proc proc, void *data)
     /*@ Allocates, initialized and adds an event entry. */
{
  pl_tevent *l = malloc(sizeof(*l));

  if ( ! l )
    return l;

  memset(l, 0, sizeof(*l));
  l->when = when;
  l->proc = proc;
  l->data2 = data;
  l->destroy = free;

  pl_tevent_q_add_event(q, l);

  return l;
}


PL_TEVENT_EXPORT int pl_tevent_q_remove(pl_tevent_q *q, void *handle)
     /*@ 
       Removes an alarm entry and resets the alarm handler.
     */
{
  pl_tevent *l = (pl_tevent*) handle;
  assert(q);
  assert(handle);

  pl_tevent_q_remove_event_1(q, l);
  pl_tevent_q_set_alarm(q);

  return 0;
}

