/* Author: kurt.stephens@parlano.com 2000/10/02 */
/*   parlano eventloop library -- pl_fd/fdobj.c
 *   Copyright (C) 2001 Parlano, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "fdobj.h"
#include "pl_el/eventloop.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h> /* inet_ntoa() */
#include <netinet/in.h> /* in_addr_t */
#ifdef linux
/*
  f***! why isnt this defining in_addr_t like the man page says?
   -- kurt.stephens@parlano.com 2000/10/09  
*/
typedef unsigned long in_addr_t;
#endif
#ifdef __CYGWIN32__
typedef unsigned long in_addr_t;
#endif

#ifndef _SOCKLEN_T
#define _SOCKLEN_T
/*
  More Solaris 5.6 .vs. 5.8 bullshit.
  -- kurt.stephens@parlano.com 2000/11/20
*/
typedef size_t socklen_t;
#endif


/**************************************************************************/
/* Util */

static char *pl_strdup(const char *x)
{
  return x ? strcpy(malloc(strlen(x) + 1), x) : 0;
}


static int sep = ':';
/*@ Separator for string socket representations. */

static const char *str_scan(char *s, size_t len, const char *r, int sep)
{
  while ( *r && -- len > 0 ) {
    if ( *r == sep ) {
      ++ r;
      break;
    }
    *(s ++) = *(r ++);
  }
  *s = 0;

  return r;
}


int _pl_str_to_sockaddr
(
 const char *name, 
 struct sockaddr_in *sa, 
 int *ossockdomainp, 
 int *ossocktypep, 
 int *osprotop, 
 char **namep)
     /*@ 
       Converts a string to a struct sockaddr_in.
       Stores the socket domain in *%ossockdomainp*, if %ossockdomainp% is not 0.
       Stores the socket type in *%ossocktypep*, if %ossocktypep% is not 0.
       Stores the socket protocol in *%osprotop*, if %osprotop% is not 0.
       Stores a cannonical name in *%namep%, if %namep% is not 0. 
     */
{
  /* WARNING: THIS CODE IS NOT THREAD-SAFE. */
  struct in_addr *ia = &(sa->sin_addr);
  char host[256];
  char service[256];
  char proto[256];
  struct hostent *he = 0;
  struct servent *se = 0;
  struct protoent *pe = 0;

  int ossockdomain = 0;
  int ossocktype = 0;
  int osproto = 0;

  assert(sa);
  assert(name);

#if 0
  fprintf(stderr, "sizeof(struct sockaddr) = %lu\n", (unsigned long) sizeof(struct sockaddr));
  fprintf(stderr, "sizeof(struct sockaddr_in) = %lu\n", (unsigned long) sizeof(struct sockaddr_in));
#endif

  memset(host, 0, sizeof(host));
  memset(service, 0, sizeof(service));
  memset(proto, 0, sizeof(proto));

  /********************************************************************/
  /* "host:service:proto" */

  {
    const char *r;

    r = name;
    r = str_scan(host, sizeof(host), r, sep);
    r = str_scan(service, sizeof(service),r, sep);
    r = str_scan(proto, sizeof(proto), r, sep);
  }

  /********************************************************************/
  /* host */

  sa->sin_family = AF_INET;

  /* Default to INADDR_ANY. */
  ia->s_addr = INADDR_ANY;
  if ( ! *host ) {
    ;
    /* Try host by name. */
  } else if ( (he = gethostbyname(host)) ) {
      ;
  } else {
    /* Try host by IP addr. */
    in_addr_t ina;

    ina = inet_addr(host);
    if ( ina != (in_addr_t) -1 ) {
      sa->sin_family = AF_INET;
      ia->s_addr = ina;
      
      he = gethostbyaddr((const char *) &ia->s_addr, sizeof(ia->s_addr), sa->sin_family);
    }
  }

  if ( he ) {
    sa->sin_family = he->h_addrtype;
    assert(sa->sin_family == AF_INET);

    assert(sizeof(sa->sin_addr) >= he->h_length);
    memcpy(&(sa->sin_addr), he->h_addr, he->h_length);

    strncpy(host, he->h_name, sizeof(host) - 1);
  }

  /********************************************************************/
  /* proto. */

  if ( ! *proto ) {
    /* Default to tcp. */
    strcpy(proto, "tcp");
  }
  else if ( (pe = getprotobyname(proto)) ) {
    /* Protocol name. */
    ;
  } 
  else if ( (pe = getprotobynumber(atoi(proto))) ) {
    /* Protocol number. */
    ;
  } 
  else {
    /* HUH? */
    osproto = atoi(proto);
  }

  if ( pe ) {
    osproto = pe->p_proto;
    strncpy(proto, pe->p_name, sizeof(proto) - 1);
  }

  /********************************************************************/
  /* service. */

  if ( ! *service ) {
    /* Default to port 0; kernel assigns an ephemeral port. */
    sa->sin_port = htons(0);
  }
  else if ( (se = getservbyname(service, proto)) ) {
    /* Service name. */
    ;
  } 
  else if ( (se = getservbyport(atoi(service), proto)) ) {
    /* Service port. */
    ;
  } 
  else {
    /* Try numeric port. */
    sa->sin_port = htons(atoi(service));
  }

  if ( se ) {
    sa->sin_port = htons(se->s_port);
    strncpy(service, se->s_name, sizeof(service) - 1);
  }

  /********************************************************************/
  /* Cannonicalize name. */

  if ( namep ) {
    *namep = realloc(*namep, strlen(host) + 1 + strlen(service) + 1 + strlen(proto) + 1);
    assert(*namep);
    sprintf(*namep, "%s%c%s%c%s", host, sep, service, sep, proto);
  }


  /********************************************************************/
  /* Return other info. */

  if ( ossockdomainp )
    *ossockdomainp = ossockdomain;

  if ( ossocktypep )
    *ossocktypep = ossocktype;

  if ( osprotop )
    *osprotop = osproto;

  return 0;
}


int _pl_sockaddr_to_str(const struct sockaddr_in *sa, int osproto, char **namep)
     /* Convert a struct sockaddr and protocol into a string. */
{
  const struct in_addr *ia = &(sa->sin_addr);
  char host[256];
  char service[256];
  char proto[256];
  struct hostent *he;
  struct servent *se;
  struct protoent *pe;
  
  assert(sa);
  assert(namep);
  /* assert(*namep); */

  memset(host, 0, sizeof(host));
  memset(service, 0, sizeof(service));
  memset(proto, 0, sizeof(proto));

  /* Get hostname for address. */
  if ( (he = gethostbyaddr((const char *)ia, sizeof(*ia), sa->sin_family)) ) {
    strcpy(host, he->h_name);
  } else {
    /* Use IP addr. */
    strncpy(host, (char*) inet_ntoa(*ia), sizeof(host) - 1);
  }

  /* Get protocol by number. */
  if ( (pe = getprotobynumber(osproto)) ) {
    strncpy(proto, pe->p_name, sizeof(proto) - 1);
  } else {
    /* Assume "tcp". */
    strcpy(proto, "tcp");
  }

  /* Get service by port and proto. */
  if ( (se = getservbyport(ntohs(sa->sin_port), proto)) ) {
    strncpy(service, se->s_name, sizeof(service) - 1);
  } else {
    /* Use port number. */
    sprintf(service, "%d", (int) (ntohs(sa->sin_port)));
  }

  /* Make space. */
  *namep = realloc(*namep, strlen(host) + 1 + strlen(service) + 1 + strlen(proto) + 1);
  assert(*namep);
  sprintf(*namep, "%s%c%s%c%s", host, sep, service, sep, proto);

  return 0;
}


/**************************************************************************/
/* configuration. */

static char *_pl_fd_local_name_default = 0;

static char *_pl_fd_local_name_default_envvar = 0;


int pl_fd_set_local_name_default(const char *x)
     /*@ Set the default local address for the socket.
       This address will be used to bind connecting sockets to a specific addression.
       This is used if %pl_fd_set_local_name% has not been called.
       If this value of the variable is null, the socket is not bound.
     */
{
  if ( _pl_fd_local_name_default )
    free(_pl_fd_local_name_default);
  _pl_fd_local_name_default = pl_strdup(x);

  return 0;
}


int pl_fd_set_local_name_default_envvar(const char *x)
     /*@ Set the name of an environment variable to get a default local address for the socket.
       The address will be used to bind connecting sockets to a specific address.
       This is used if %pl_fd_set_local_name% and %pl_fd_set_local_name_default% have not been called.
       If this value of the variable is null, the socket is not bound.
     */
{
  if ( _pl_fd_local_name_default_envvar )
    free(_pl_fd_local_name_default_envvar);
  _pl_fd_local_name_default_envvar = pl_strdup(x);

  return 0;
}


const char *pl_fd_local_name_default()
     /*@
       Returns the default local address to bind connecting sockets.
       This is used if %pl_fd_set_local_name%() is not called.
     */
{
  const char *x = 0;

  do {
    /* Try explicit default local address. */
    if ( _pl_fd_local_name_default && *_pl_fd_local_name_default ) {
      x = _pl_fd_local_name_default;
      break;
    }
    /* Try default local address in environment variable. */
    if ( _pl_fd_local_name_default_envvar && *_pl_fd_local_name_default_envvar ) {
      x = getenv(_pl_fd_local_name_default_envvar);
      if ( x && *x )
	break;
    }
  } while ( 0 );

  return x;
}





/**************************************************************************/
/* High-level I/O */

int pl_fd_read(pl_fd *fdo, void *buf, size_t size)
     /*@ 
       Attempts to read %size% bytes from %fdo% into %buf%
       Returns actual number of bytes read.
       Returns 0 if eof.
       Returns -1 if error.
       Returns -2 if %fdo% is non-blocking and a block would have occurred.
     */
{
  assert(fdo);
  assert(fdo->osread);

  return fdo->osread(fdo, buf, size);
}


int pl_fd_write(pl_fd *fdo, const void *buf, size_t size)
     /*@ 
       Attempts to write %size% bytes from %buf% to %fdo%
       Returns actual number of bytes written.
       Returns 0 if eof.
       Returns -1 if error.
       Returns -2 if %fdo% is non-blocking and a block would have occurred.
     */
{
  assert(fdo);
  assert(fdo->oswrite);

  return fdo->oswrite(fdo, buf, size);
}


static pl_el_callback_handler_DECL(_pl_fd_retry_open)
     /*@ Called when after %fdo%->retry_delay seconds after failed open. */
{
  pl_fd *fdo = data;
  int result = 0;

  /* Timer triggered. */
  fdo->retry_open_timer = 0;

  if ( fdo->event_retry_open ) {    
    result = fdo->event_retry_open(fdo, el);
  }
  
  if ( result == 0 ) {
    /* Attempt open again. */
    result = pl_fd_open(fdo);
  }

  return 0;
}


int pl_fd_open(pl_fd *fdo)
     /*@ 
       Opens %fdo% for I/O.
       If the "open" fails and %fdo%->retry_delay is non-zero,
       an open attempt is tried %fdo%->retry_max every %fdo%->retry_delay seconds.
       %fdo%->opened is called upon sucess if %fdo%->opened is non-null.
       Returns < 0 if error.
     */
{
  int result;

  assert(fdo);
  assert(fdo->osopen);

  result = fdo->osopen(fdo);

  /* Attempt async retry? */
  if ( result ) {
    /* Retry enabled? */
    if ( fdo->retry_open_max ) {
      /* Too many retries? */
      if ( fdo->retry_open_max < 0 || fdo->retry_open_count < fdo->retry_open_max ) {
	++ fdo->retry_open_count;
	if ( fdo->retry_open_timer == 0 ) {
	  /* Create timer to reopen after retry_delay seconds. */
	  pl_tevent_time now = pl_tevent_time_now();
	  pl_tevent_time offset = pl_time_t_to_tevent_time(fdo->retry_open_delay);
	  pl_tevent_time when = pl_tevent_time_add(now, offset);

	  fdo->retry_open_timer = pl_el_at_time(fdo->el, when, _pl_fd_retry_open, fdo);
	  if ( fdo->retry_open_timer ) {
	    /* Timer created. */

	    /* Disable events. */
	    fdo->event_enable = 0;

	    result = 0;
	  } else {
	    /* Timer created failed. */
	    result = -1;
	  }
	}
      }
      
      /* Retry failed. */
      if ( result != 0 ) {
	if ( fdo->event_retry_open_error ) {
	  result = fdo->event_retry_open_error(fdo, fdo->el);
	}
      }

    }
  } else {
    /* Renabled events. */
    fdo->event_enable = 1;

    /* Notify. */
    if ( fdo->event_opened ) {
      result = fdo->event_opened(fdo, fdo->el);
    }

    /* Reset count. */
    fdo->retry_open_count = 0;
  }
  
      
  return result;
}


int pl_fd_close_(pl_fd *fdo)
{
  int result = 0;

  if ( fdo->osclose ) {
    result = fdo->osclose(fdo);
  }
  
  if ( result == 0 ) {
    fdo->fd = pl_osfd_INVALID;
    fdo->state = pl_fd_CLOSED;
    fdo->event_enable = 0;
  }

  return result;
}


int pl_fd_close(pl_fd *fdo)
     /*@ 
       Closes %fdo% for I/O.
       %fdo%->closing is called if %fdo%->closing is non-null.
       Returns < 0 if error.
     */
{
  int result = 0;

  assert(fdo);

  do {
    if ( fdo->state >= pl_fd_OPENED ) {
      if ( fdo->event_closing && (result = fdo->event_closing(fdo, fdo->el)) )
	break;

      result = pl_fd_close_(fdo);
    }
  } while(0);

  return result;
}


int pl_fd_reopen(pl_fd *fdo)
     /*@ 
       Closes and reopens %fdo%.
       Returns < 0 if error.
     */
{
  assert(fdo);

  return pl_fd_close(fdo) || pl_fd_open(fdo);
}


/**************************************************************************/
/* Wrap POSIX system calls. */

#ifndef PLFDO_DEBUG
#define PLFDO_DEBUG ((fdo)->debug)
#else
#define PLFDO_DEBUG 9
#endif

#ifndef EWOULDBLOCK
#define EWOULDBLOCK 0
#endif
#ifndef EAGAIN
#define EAGAIN 0
#endif
#ifndef EINPROGRESS
#define EINPROGRESS 0
#endif
#ifndef EALREADY
#define EALREADY 0
#endif

#define pl_fd_SYSCALL(RESULT,FDO,CALL) do { \
  do {  \
    (RESULT) = CALL; \
    if ( PLFDO_DEBUG > 2 ) fprintf(stderr, "%s: %s = %d\n", (FDO)->name, #CALL, (RESULT)); \
  } while ( (RESULT) < 0 && errno == EINTR ); \
  if ( (RESULT) < 0 ) { \
    if ( (EWOULDBLOCK && errno == EWOULDBLOCK) || (EAGAIN && errno == EAGAIN) ) { \
      if ( PLFDO_DEBUG > 1 ) fprintf(stderr, "%s: EWOULDBLOCK\n", #CALL); \
      (RESULT) = pl_fd_EWOULDBLOCK; \
    } else if ( EINPROGRESS && errno == EINPROGRESS ) { \
      if ( PLFDO_DEBUG > 1 ) fprintf(stderr, "%s: EINPROGRESS\n", #CALL); \
      (RESULT) = pl_fd_EINPROGRESS; \
    } else if ( EALREADY && errno == EALREADY ) { \
      if ( PLFDO_DEBUG > 1) fprintf(stderr, "%s: EALREADY\n", #CALL); \
      (RESULT) = pl_fd_EALREADY; \
    } else { \
      (FDO)->oserr = errno; \
      (FDO)->oserrstr = 0; \
    } \
  } \
} while (0)


/**************************************************************************/
/* POSIX methods. */

int _pl_fd_write(pl_fd *fdo, const void *buf, size_t size)
{
  int result;
  
  pl_fd_SYSCALL(result, fdo, write(fdo->fd, buf, size));

  if ( fdo->debug ) {
    fprintf(stderr, "%s: write(%d, %p, %lu) = %ld: %s\n", fdo->name, fdo->fd, buf, size, result, pl_fd_get_oserrstr(fdo));
  }

  return result;
}


int _pl_fd_read(pl_fd *fdo, void *buf, size_t size)
{
  int result;
  
  pl_fd_SYSCALL(result, fdo, read(fdo->fd, buf, size));

  if ( fdo->debug ) {
    fprintf(stderr, "%s: read(%d, %p, %lu) = %ld: %s\n", fdo->name, fdo->fd, buf, size, result, pl_fd_get_oserrstr(fdo));
  }

  return result;
}


int _pl_fd_close(pl_fd *fdo)
{
  int result;
  
  pl_fd_SYSCALL(result, fdo, close(fdo->fd));

  return result;
}


int _pl_fd_shutdown(pl_fd *fdo)
{
  int result;
  
  pl_fd_SYSCALL(result, fdo, shutdown(fdo->fd, 2)); /* 2: no read or writes. */

  return result;
}


/********************************************/
/* State. */

pl_fd_state pl_fd_get_state(pl_fd *fdo)
{
  assert(fdo);

  return fdo->state;
}


int pl_fd_get_oserr(pl_fd *fdo)
{
  assert(fdo);

  return fdo->oserr;
}


const char *pl_fd_get_oserrstr(pl_fd *fdo)
{
  assert(fdo);

  /* Cache the strerror() call. */
  if ( ! fdo->oserrstr ) {
    fdo->oserrstr = strerror(fdo->oserr);
  }

  return fdo->oserrstr;
}



/**************************************************************************/

int pl_fd_init(pl_fd *fdo)
     /*@ 
       Initializes %fdo% for subsequent %pl_fd_file%, %pl_fd_socket% call.
       Returns non-zero if error.
     */
{
  assert(fdo);

  assert(sizeof(int) == sizeof(size_t));

  memset(fdo, 0, sizeof(*fdo));

  fdo->fd = pl_osfd_INVALID;
  fdo->state = pl_fd_INVALID;
  fdo->osread = _pl_fd_read;
  fdo->oswrite = _pl_fd_write;
  fdo->osclose = _pl_fd_close;
  fdo->osshutdown = fdo->osclose;

  fdo->name = malloc(1);
  assert(fdo->name);
  fdo->name[0] = 0;

#if 0
  fdo->local_name = malloc(1);
  assert(fdo->local_name);
  fdo->local_name[0] = 0;
#endif

  return 0;
}


int pl_fd_destroy(pl_fd *fdo)
     /*@ 
       Closes and destroys %fdo%.
       %fdo%->destroy(%fdo%) is called if %fdo%->destroy is non-null.
       %fdo% is removed from its current event loop, if any.
       %fdo% is closed, if open.
       %fdo%->dealloc(%fdo%) is called if %fdo%->decalloc is non-null.
       Returns non-zero if error.
     */
{
  int result = 0;

  assert(fdo);

  do {
    int save_state = fdo->state;

    /* Destroy called recursively. */
    if ( fdo->state <= pl_fd_DESTROYED ) {
      result = -1;
      break;
    }

    /* Mark this as destroyed. */
    fdo->state = pl_fd_DESTROYED;

    /* Call user destroy method. */
    if ( fdo->destroy ) {
      /* Call fdo->destroy only once. */
      int (*x)(pl_fd *) = fdo->destroy;
      if ( (result = x(fdo)) ) {
	fdo->state = save_state;
	break;
      }
      fdo->destroy = 0;
    }

    /* Remove from current event loop. */
    if ( fdo->el ) {
      pl_el_remove(fdo->el, fdo);
      fdo->el = 0;
    }

    /* Temporarily restore state so pl_fd_close() works. */
    fdo->state = save_state;
    if ( fdo->state >= pl_fd_HAS_FD ) {
      result = pl_fd_close(fdo);
    }
    fdo->state = pl_fd_DESTROYED;

    /* Free name. */
    free(fdo->name);
    fdo->name = 0;

    /* Free local name. */
    free(fdo->local_name);
    fdo->local_name = 0;
    
    /* Call user deallocation method. */
    if ( fdo->dealloc ) {
      /* Call fdo->dealloc only once. */
      void (*x)(void *ptr) = fdo->dealloc;
      fdo->dealloc = 0;
      x(fdo);
    }
  } while ( 0 );

  return result;
}



/**************************************************************************/
/* POSIX fcntl() */


int _pl_fd_fcntl_getflag(pl_fd *fdo, int flag)
{
  int flags;

#if defined(F_GETFL)
  pl_fd_SYSCALL(flags, fdo, fcntl(fdo->fd, F_GETFL, 0));
  if ( flags == -1 ) {
    return -1;
  }
#else
  return -1;
#endif

  fdo->osflags = flags;

  return flags & flag ? 1 : 0;
}


int _pl_fd_fcntl_setflag(pl_fd *fdo, int flag, int set)
{
  int result;
  int flags = 0;
  
  assert(fdo);

#if defined(F_GETFL)
  pl_fd_SYSCALL(result, fdo, fcntl(fdo->fd, F_GETFL, 0));
  if ( result == -1 ) {
    return -1;
  }
  flags = result;
#else
  return -1;
#endif

  if ( set ) {
    flags |= flag;
  } else {
    flags &= ~ flag;
  }

#if defined(F_SETFL)
  pl_fd_SYSCALL(result, fdo, fcntl(fdo->fd, F_SETFL, flags));
  if ( result == -1 ) {
    return -1;
  }
#else
  return -1;
#endif

  fdo->osflags = flags;
  
  return 0;
}


/**************************************************************************/
/* High-level options. */


int pl_fd_nblock(pl_fd *fdo)
     /*@ 
       Gets non-blocking mode on %fdo%.
       Returns -1 if error.
     */
{
#if defined(O_NONBLOCK)
  if ( pl_osfd_IS_VALID(fdo->fd) ) {
    fdo->nblock = _pl_fd_fcntl_getflag(fdo, O_NONBLOCK);
  }
  return fdo->nblock; 
#else
  fdo->nblock = 0;
  return -1;
#endif
}


int pl_fd_set_nblock(pl_fd *fdo, int nblock)
     /*@ 
       Sets non-blocking mode on %fdo%.
       Returns -1 if error.
     */
{
  int result = -1;

#if defined(O_NONBLOCK)
  fdo->nblock = nblock;
  if ( pl_osfd_IS_VALID(fdo->fd) ) {
    result = _pl_fd_fcntl_setflag(fdo, O_NONBLOCK, nblock);
  } else {
    /* Deferred until we have a valid fd. */
    result = 0;
  }
#endif

  return result;
}



/**************************************************************************/
/* POSIX file descriptor method. */


int _pl_fd_fd_open(pl_fd *fdo)
{
  assert(fdo);

  return 0;
}


int _pl_fd_fd_close(pl_fd *fdo)
{
  assert(fdo);

  return 0;
}


int pl_fd_fd(pl_fd *fdo, pl_osfd fd)
     /*@ 
       Initializes %fdo% from a raw fd.
       The fd is not closed on %pl_fd_close%.
       Returns non-zero if error.
     */
{
  assert(fdo);

  /* Save name. */
  fdo->name = realloc(fdo->name, 32);
  assert(fdo->name);
  sprintf(fdo->name, "/dev/fd/%d", (int) fd);
  fdo->name = realloc(fdo->name, strlen(fdo->name) + 1);
  assert(fdo->name);

  /* Save fd. */
  fdo->fd = fd;
  fdo->state = pl_fd_HAS_FD;

  /* Get non-blocking option. */
  pl_fd_nblock(fdo);

  /* open and close methods do nothing. */
  fdo->osopen = _pl_fd_fd_open;
  fdo->osclose = _pl_fd_fd_close;
  fdo->osshutdown = fdo->osclose;

  return 0;
}


/**************************************************************************/
/* POSIX file open method. */

int _pl_fd_file_open(pl_fd *fdo)
{
  int result = -1;

  assert(fdo);

  /* open file now. */
  pl_fd_SYSCALL(result, fdo, open(fdo->name, fdo->osmode, fdo->oscreat));

  if ( result >= 0 ) {
    fdo->fd = result;
    fdo->state = pl_fd_OPENED;

    /* Set non-blocking option. */
    pl_fd_set_nblock(fdo, fdo->nblock);

    result = 0;

    if ( fdo->event_opened ) {
      result = fdo->event_opened(fdo, fdo->el);
    }
  } else {
    result = -1;
  }

  return result;
}



int pl_fd_file(pl_fd *fdo, const char *file, int mode, int create)
     /*@ 
       Initializes %fdo% for reading or writing to %file%.
       Read is enabled if %mode% & 4.
       Write is enabled if %mode% & 2.
       File is created if %create% is non-zero.
       Returns non-zero if error.
     */
{
  assert(fdo);

  fdo->osmode = 0;
  fdo->oscreat = create;

  assert(O_RDWR == (O_RDONLY | O_WRONLY));
  if ( mode & 0111 ) fdo->osmode |= O_TRUNC;
  if ( mode & 0444 ) fdo->osmode |= O_RDONLY;
  if ( mode & 0222 ) fdo->osmode |= O_WRONLY;
  if ( fdo->osmode & O_WRONLY && ! (fdo->osmode & O_TRUNC) ) fdo->osmode |= O_CREAT;
  
  if ( fdo->oscreat == 0 && fdo->osmode & O_CREAT) fdo->oscreat = 0600;

  fdo->name = realloc(fdo->name, strlen(file) + 1);
  assert(fdo->name);
  strcpy(fdo->name, file);

  fdo->osopen = _pl_fd_file_open;

  return pl_fd_open(fdo);
}


/**************************************************************************/
/* POSIX socket support. */

#ifndef MSG_NOSIGNAL
#define MSG_NOSIGNAL 0
#endif

int _pl_fd_send(pl_fd *fdo, const void *buf, size_t size)
{
  int result;
  socklen_t sl = sizeof(fdo->os_send_addr);
 
  pl_fd_SYSCALL(result, fdo, sendto(fdo->fd, buf, size, MSG_NOSIGNAL, (struct sockaddr*) &fdo->os_send_addr, sl));

  return result;
}


int _pl_fd_recv(pl_fd *fdo, void *buf, size_t size)
{
  int result;
  
  socklen_t sl = sizeof(fdo->os_recv_addr);

  pl_fd_SYSCALL(result, fdo, recvfrom(fdo->fd, buf, size, MSG_NOSIGNAL, (struct sockaddr*) &fdo->os_recv_addr, &sl));

  return result;
}



const char *pl_fd_local_name(pl_fd *fdo)
     /* Get the local address for the client socket.
      Defaults to %pl_fd_local_name_default%() if %fdo%->local_name is 0.
     */
{
  assert(fdo);

  return (fdo->local_name) ? 
    fdo->local_name : 
    pl_fd_local_name_default();
}


int pl_fd_set_local_name(pl_fd *fdo, const char *name)
     /*@ Set the local address for the client socket. 
      This address is used to bind the client socket to a specific address instead of the OS default.
      You should set this if you are used multiple network interfaces.
     */
{
  assert(fdo);

  free(fdo->local_name);
  fdo->local_name = pl_strdup(name);

  return 0;
}


static int _pl_fd_sockaddr_to_name(pl_fd *fdo)
{
  return _pl_sockaddr_to_str(&(fdo->ossockaddr), fdo->osproto, &(fdo->name));
}


static int _pl_fd_sockaddr_from_name(pl_fd *fdo)
{
  int result = -1;

  do {
    const char *x;

    if ( (result = _pl_str_to_sockaddr(fdo->name, &(fdo->ossockaddr), 0, 0, &(fdo->osproto), &(fdo->name))) )
      break;
    
    /* Save for send() and recv(). */
    fdo->os_recv_addr = fdo->ossockaddr;
    fdo->os_send_addr = fdo->ossockaddr;
    fdo->os_local_sockaddr = fdo->ossockaddr;
   
  } while ( 0 );

  return result;
}



int _pl_fd_socket(pl_fd *fdo)
{
  int result = -1;

  assert(fdo);

  if ( fdo->state >= pl_fd_HAS_FD )
    return 0;

  /* Protocol family. */
  fdo->ossockdomain = PF_INET;

  /* Protocol type. */
  switch ( fdo->osproto ) {
  case IPPROTO_TCP:
    /* TCP/IP is a stream protocol. */
    fdo->ossocktype = SOCK_STREAM;
    break;

  case IPPROTO_UDP:
    /* UDP is a datagram protocol. */
    fdo->ossocktype = SOCK_DGRAM;
#if 0
    /* Use recv and send for all datagram socket IO. */
    fdo->read = _pl_fd_recv;
    fdo->write = _pl_fd_send;
#endif
    break;

  default:
    /* ??? */
    assert(! "TCP || UDP");
    break;
  }

  pl_fd_SYSCALL(result, fdo, socket(fdo->ossockdomain, fdo->ossocktype, fdo->osproto));

  if ( result >= 0 ) {
    fdo->fd = result;
    fdo->state = pl_fd_HAS_FD;

    /* Reset nblock status. */
    pl_fd_set_nblock(fdo, fdo->nblock);

#ifdef SOL_SOCKET
#ifdef SO_REUSEADDR
    {
      int opt = 1;
      pl_fd_SYSCALL(result, fdo, setsockopt(fdo->fd, SOL_SOCKET, SO_REUSEADDR, (const char*) &opt, sizeof(opt)));
    }
#endif

#ifdef SO_LINGER
    {
      struct linger opt = { 0, 0 };
      pl_fd_SYSCALL(result, fdo, setsockopt(fdo->fd, SOL_SOCKET, SO_LINGER, (const char*) &opt, sizeof(opt)));
    }
#endif
#endif
  }
    
  return result;
}



static int _pl_fd_bind(pl_fd *fdo)
{
  int result = -1;

  if ( fdo->state < pl_fd_BOUND ) {
    /* Bind socket to accept from address. */
    pl_fd_SYSCALL(result, fdo, bind(fdo->fd, (const struct sockaddr *) &fdo->os_local_sockaddr, sizeof(fdo->ossockaddr))); 

    if ( result >= 0 ) {
      fdo->state = pl_fd_BOUND;
    }
  } else {
    /* Bound already done. */
    result = 0;
  }

  return result;
}


/**********************************************************************/
/* Server side sockets. */



static int _pl_fd_listen(pl_fd *fdo)
{
  int result = -1;

  assert(fdo);

  if ( fdo->state < pl_fd_BOUND ) {
    /* Bind socket to accept from address. */
    fdo->os_local_sockaddr = fdo->ossockaddr;

    result = _pl_fd_bind(fdo);
    if ( result >= 0 ) {
      fdo->state = pl_fd_BOUND;
    }
  } else {
    /* Bound already done. */
    result = 0;
  }
  

  if ( result >= 0 ) {
    switch ( fdo->ossocktype ) {
    case SOCK_STREAM:
    case SOCK_SEQPACKET:
      if ( fdo->state == pl_fd_BOUND ) {
	/* TCP needs listen. */
	pl_fd_SYSCALL(result, fdo, listen(fdo->fd, SOMAXCONN));

	if ( result >= 0 ) {
	  fdo->state = pl_fd_LISTENING;
	}
      } else {
	result = -1;
      }
      break;

    case SOCK_DGRAM:
      fdo->state = pl_fd_LISTENING;
      break;

    default:
      /* ??? */
      assert(! "SOCK_STREAM | SOCK_SEQPACKET | SOCK_DGRAM");
      break;
    }
  }

  return result;
}


int _pl_fd_socket_server_open(pl_fd *fdo)
{
  int result = -1;

  assert(fdo);
   
  if ( (result = _pl_fd_sockaddr_from_name(fdo)) )
    return result;

  if ( (result = _pl_fd_socket(fdo)) < 0 )
    return result;

  /* Default local address from ossockaddr. */
  fdo->os_local_sockaddr = fdo->ossockaddr;
    
  if ( (result = _pl_fd_listen(fdo)) < 0 )
    return result;

  return result;
}


int pl_fd_init_proto(pl_fd *fdo, pl_fd *proto)
     /* Initializes %fdo% from %proto%. */
{
  int result = 0;

  assert(fdo);
  assert(proto);

  fdo->async = proto->async;
  fdo->nblock = proto->nblock;
  fdo->osmode = proto->osmode;
  fdo->oscreat = proto->oscreat;
  fdo->proto = proto;
  fdo->event_enable = proto->event_enable;
  fdo->event_mask = proto->event_mask & ~(pl_fd_OSREAD|pl_fd_OSWRITE|pl_fd_OSEXCEPT);
  fdo->event_read = proto->event_read;
  fdo->event_write = proto->event_write;
  fdo->event_except = proto->event_except;
  fdo->event_accepting = proto->event_accepting;
  fdo->event_accepted = proto->event_accepted;
  fdo->event_connecting = proto->event_connecting;
  fdo->event_opened = proto->event_opened;
  fdo->event_retry_open = proto->event_retry_open;
  fdo->event_retry_open_error = proto->event_retry_open_error;
  fdo->retry_open_max = proto->retry_open_max;
  fdo->retry_open_delay = proto->retry_open_delay;
  fdo->destroy = proto->destroy;
  fdo->debug = proto->debug;
  memcpy(&fdo->opaque, &proto->opaque, sizeof(fdo->opaque)); 

  return result;
}


int pl_fd_socket_server(pl_fd *fdo, const char *name)
     /*@ 
       Initializes %fdo% as a bound server-side socket.
       %name% should contain three fields separated by a ':' character: "host:service:protocol".
       if host is null, all client's are allowed to connect.
       host may be a dotted-decimal IP address or a host name.
       service may be a decimal port number or a service name.
       protocol may be a decimal protocol number or a protocol name.
       if protocol is null, "tcp" is used.
       Returns non-zero if error.
     */
{
  assert(fdo);
  
  /* Save name. */
  free(fdo->name);
  fdo->name = pl_strdup(name);

  /* Set open method. */
  fdo->osopen = _pl_fd_socket_server_open;

  return 0;
}



static pl_fd_event_handler_DECL(_pl_fd_srv_event_accept)
     /*@ Called by event loop when server-side %fdo% can accept a client. */
{
  int result;
  pl_fd *fdoc;

  do {
    /* Allocate a new accepted socket. */
    fdoc = malloc(sizeof(*fdoc));
    if ( ! fdoc ) {
      result = -1;
      break;
    }
    
    /* Initialize it. */
    if ( (result = pl_fd_init(fdoc)) )
      break;

    /* Use event handlers from server's proto. */
    if ( fdo->proto ) {
      pl_fd_init_proto(fdoc, fdo->proto);
    }

    /* Was malloc'ed. */
    fdoc->dealloc = pl_delayed_free;

    /* Get ready for accept. */
    if ( (result = pl_fd_accept_async(fdoc, fdo)) )
      break;

    /* Start accept sequence. */
    if ( (result = pl_fd_open(fdoc)) )
      break;

    /* Add to event loop. */
    pl_el_add(el, fdoc);

  } while ( 0 );

  if ( result ) {
    pl_fd_destroy(fdoc);
  }

  return result;
}


int _pl_fd_socket_server_async_open(pl_fd *fdo)
{
  int result = -1;

  do {
    if ( (result = _pl_fd_socket_server_open(fdo)) )
      break;
   
    /* Read events for listening sockets means that a client is connecting. */
    fdo->event_enable = 1;
    fdo->event_mask |= pl_fd_OSREAD | pl_fd_OSWRITE | pl_fd_OSEXCEPT;
    fdo->event_oswrite =
      fdo->event_osread = _pl_fd_srv_event_accept;
 
  } while ( 0 );

  return result;
}



int pl_fd_socket_server_async(pl_fd *fdo, const char *name, pl_fd *proto)
     /*@ 
       Initializes %fdo% as a bound server-side socket that asynchronously accepts client connections.
       When a client starts connecting, %proto% is used to initialize a new client connection object.
     */
 {
   int result = -1;

   do {
     if ( (result = pl_fd_socket_server(fdo, name)) )
       break;

     fdo->proto = proto;
     fdo->async = 1;
     fdo->osopen = _pl_fd_socket_server_async_open;

   } while ( 0 );
   
   return result;
}


/**********************************************************************/
/* Accepted server-side sockets. */


int _pl_fd_accept_open(pl_fd *fdo)
     /*@
       Called be pl_fd_open() for pl_fd_accept() sockets.
     */
{
  int result = -1;
  struct sockaddr sa;
  socklen_t socklen;
  int nblock_save = 0;

  pl_fd_event_handler callback0 = 0, callback1 = 0;

  pl_fd *fdos = fdo->accept_from;

  assert(fdo->accept_from);

  do {
    switch ( fdo->ossocktype ) {
      /* Only these types of sockets need accept() */
    case SOCK_STREAM:
    case SOCK_SEQPACKET:
    case SOCK_RDM:
      socklen = sizeof(sa);
      assert(sizeof(sa) >= sizeof(fdos->ossockaddr));
      memcpy(&sa, &(fdos->ossockaddr), socklen);
      
      /* Only listening or accepting sockets should call accept(). */
      if ( fdo->accept_from->state != pl_fd_LISTENING ) {
	result = -1;
	break;
      }
      
      /*
	Note: fdo->fd is the bound fd, initialized in pl_fd_accept. 
	This is so we can use an event loop to tell us when the accept() can complete.
      */

      /* If async, set nblock temporarily for POSIX accept. */
      if ( fdo->async ) {
	nblock_save = pl_fd_nblock(fdo);
        pl_fd_set_nblock(fdo, 1);
      }

      pl_fd_SYSCALL(result, fdo, accept(fdo->fd, (struct sockaddr *) &sa, &socklen));
      
      if ( fdo->async ) {
	pl_fd_set_nblock(fdo, nblock_save);
      }

      /* already started accept. */
      if ( result == pl_fd_EALREADY ) {
	if ( fdo->state != pl_fd_ACCEPTING ) {
	  fdo->accept_from->state = 
	    fdo->state = pl_fd_ACCEPTING;

	  callback0 = fdo->event_accepting;
	}
	result = 0;
	break;
      }
      else if ( result == pl_fd_EINPROGRESS ) {
	/* fdo->fd was O_NONBLOCK, started handshake. */
	if ( fdo->state != pl_fd_ACCEPTING ) {
	  fdo->accept_from->state = 
	    fdo->state = pl_fd_ACCEPTING;
	
	  callback0 = fdo->event_accepting;
	}
	break;
      } 
      else if ( result >= 0 ) {
	/* Handshake is done. */
	/* New fd was accepted(). */
	if ( fdo->state != pl_fd_ACCEPTING ) {
	  callback0 = fdo->event_accepting;
	}
	if ( fdo->state != pl_fd_ACCEPTED ) {
	  fdo->state = pl_fd_ACCEPTED;
	  
	  callback1 = fdo->event_accepted;
	}

	/* Bound socket is listening again. */
	fdo->accept_from->state = pl_fd_LISTENING;
      } else {
	/* Some other error occurred. */
	fdo->state = pl_fd_HAS_FD;
	fdo->accept_from->state = pl_fd_LISTENING;
	
	callback0 = fdo->event_accept_error;
      }
      
      if ( result >= 0 ) {
	fdo->fd = result;
	fdo->oserr = 0;
	fdo->oserrstr = 0;
	
	/* fdo-fd is a new socket. */
	pl_fd_set_nblock(fdo, fdo->nblock);
	
	fdo->osopen = 0; /* Open does not make sense more than once. */
	fdo->osread = fdos->osread;
	fdo->oswrite = fdos->oswrite;
	fdo->osclose = fdos->osclose;
	fdo->osshutdown = fdos->osshutdown;
	
	/* Set source name for sockaddr. */
	assert(sizeof(fdo->ossockaddr) >= socklen);
	memcpy(&(fdo->ossockaddr), &sa, socklen);
	_pl_fd_sockaddr_to_name(fdo);
	
	result = 0;
      }
      else if ( result == -1 ) {
	fdo->oserr = errno;
	fdo->oserrstr = 0;
	
	callback0 = fdo->event_accept_error;
      }
      break;
      
    default:
      /* Socket type does not need accept(). */
      result = -1;
      break;
    }

  } while ( 0 );

  /* Notify. */
  do {
    if ( callback0 && (result = callback0(fdo, fdo->el)) )
      break;

    if ( callback1 && (result = callback1(fdo, fdo->el)) )
      break;
  } while(0);


  return result;
}


int pl_fd_accept(pl_fd *fdo, pl_fd *fdos)
     /*@ 
       Prepares %fdo% as a client to be accept()'ed from server-side socket %fdos%.
       Must call %pl_fd_open%(%fdo%) to initiate accept().
       Returns non-zero if error.
     */
{
  int result = 0;
 
  assert(fdo);
  assert(fdos);

  switch ( fdos->ossocktype ) {
    /* Only these types of sockets need accept() */
  case SOCK_STREAM:
  case SOCK_SEQPACKET:
  case SOCK_RDM:
    fdo->accept_from = fdos;
    fdo->fd = fdos->fd;
    fdo->state = pl_fd_HAS_FD;
    
#if 0
    /* Should this really enherit non-blocking? */
    fdo->nblock = fdos->nblock;
#endif

    fdo->oserr = 0;
    fdo->oserrstr = 0;
    
    fdo->ossockdomain = fdos->ossockdomain;
    fdo->ossocktype = fdos->ossocktype;
    fdo->osproto = fdos->osproto;

    /* Allow nothing but accept, for now. */
    fdo->osread = 0;
    fdo->oswrite = 0;
    fdo->osopen = _pl_fd_accept_open;
    fdo->osclose = 0;
    fdo->osshutdown = 0;
    break;

  default:
    /* invalid socket type. */
    result = -1;
    break;
  }

  return result >= 0 ? 0 : -1;
}


static pl_fd_event_handler_DECL(_pl_fd_cli_event_accept)
{
  int result;

  result = pl_fd_open(fdo);

  return result;
}


int _pl_fd_accept_async_open(pl_fd *fdo)
{
  int result;

  do {
    /* Use _pl_fd_accept open method. */
    result = _pl_fd_accept_open(fdo);

    switch ( fdo->state ) {
    case pl_fd_ACCEPTING:
      fdo->event_enable = 1;
      fdo->event_mask |= pl_fd_OSREAD | pl_fd_OSWRITE;
      fdo->event_osread = 
      fdo->event_oswrite = _pl_fd_cli_event_accept;

      /* Disable events in bound socket until accepted. */
      assert(fdo->accept_from);
      if ( fdo->accept_from ) {
	fdo->accept_from->state = pl_fd_ACCEPTING;
	fdo->accept_from->event_enable = 0;
      }
      break;

    case pl_fd_ACCEPTED:
    default:
      if ( fdo->event_osread == _pl_fd_cli_event_accept ) {
	fdo->event_osread = 0;
	fdo->event_mask &= ~ pl_fd_OSREAD;
      }
      if ( fdo->event_oswrite == _pl_fd_cli_event_accept ) {
	fdo->event_oswrite = 0;
	fdo->event_mask &= ~ pl_fd_OSWRITE;
      }

      /* Reenable events in bound socket. */
      assert(fdo->accept_from);
      if ( fdo->accept_from ) {
	fdo->accept_from->state = pl_fd_LISTENING;
	fdo->accept_from->event_enable = 1;
      }
      break;
    }

  } while ( 0 );
  return result;
}


int pl_fd_accept_async(pl_fd *fdo, pl_fd *fdos)
     /*@
       Accepts %fdo% from bound %fdos% asynchonously. 
       Calls %fdo%->accepting when accept().
       Calls %fdo%->accepted when %fdo% has been accepted.
      */
{
  int result;

  do {
    if ( (result = pl_fd_accept(fdo, fdos)) )
      break;

    fdo->async = 1;
    fdo->osopen = _pl_fd_accept_async_open;
    
  } while ( 0 );

  return result;
}


/**********************************************************************/
/* Client side sockets. */

/* the behaviour for this function was lifted out of two places:  the "default" 
 * case behaviour for _pl_fd_connect() and the retry enqueue behaviour for 
 * pl_fd_open().  it really belongs in the 
 */
static pl_fd_event_handler_DECL(_pl_fd_cli_connect_inprogress_except)
     /*@
       Called when an inprogress connecting client socket 
       receives an exception.
     */
{
    int so_error = 0, so_error_len = 0, result = 0;

    so_error_len = sizeof(so_error);

    /* XXX:  HACK! HACK! HACK! restore the original handler */
    if ( 0 != fdo->__save_event_except ) {
	fdo->event_except = fdo->__save_event_except;
	fdo->__save_event_except = 0;
    }

    if ( pl_fd_CONNECTING != fdo->state ) {
	if ( fdo->debug ) {
	    fprintf(stderr, "%s: except during connect: fdo not in CONNECTING\n", fdo->name);
	}
	assert( pl_fd_CONNECTING == fdo->state );
	return 1;
    }
    if ( 0 > getsockopt(fdo->fd, SOL_SOCKET, SO_ERROR, &so_error, &so_error_len) ) {
	/* getsockopt errored out */
	if ( fdo->debug ) {
	    fprintf(stderr, "%s: except during connect: getsockopt failed %d: %s\n", fdo->name, errno, strerror(errno));
	}
	result = 1;
    } else {
	if (so_error) {
	    if ( fdo->debug ) {
		fprintf(stderr, "%s: except during connect: error %d: %s\n", fdo->name, so_error, strerror(so_error));
	    }
	    /* connect failed */
	    pl_fd_close_(fdo);

	    if (fdo->event_connect_error) {
		result = (fdo->event_connect_error)(fdo, fdo->el);
		if ( ! result) {
		    /* check and attempt a reconnect setup */
		    if ( fdo->retry_open_max ) {
			/* too many retries? */
			if ( fdo->retry_open_max < 0 || fdo->retry_open_count < fdo->retry_open_max ) {
			    ++ fdo->retry_open_count;
			    if ( 0 == fdo->retry_open_timer ) {
				/* create timer to reopen after retry_delay seconds. */
				pl_tevent_time now = pl_tevent_time_now();
				pl_tevent_time offset = pl_time_t_to_tevent_time(fdo->retry_open_delay);
				pl_tevent_time when = pl_tevent_time_add(now, offset);

				fdo->retry_open_timer = pl_el_at_time(fdo->el, when, _pl_fd_retry_open, fdo);
				if ( fdo->retry_open_timer ) {
				    /* timer created. */
				    
				    /* disable events. */
				    fdo->event_enable = 0;

				    result = 0;
				} else {
				    /* timer creation failed. */
				    result = -1;
				}
			    }
			}

			if ( 0 != result ) {
			    if ( fdo->event_retry_open_error ) {
				result = fdo->event_retry_open_error(fdo, fdo->el);
			    }
			}
		    }
		}
	    }   
	} else {
	    /* there is no error on the socket now.  how did we get here? */
	    if ( fdo->debug ) {
		fprintf(stderr, "%s: except during connect: no error on socket.  moving to connected.\n", fdo->name);
	    }

	    /* XXX: are we _sure_ it's connected at this point?  and do we need to do something
	     *      else like enable the events?  (i thought that a valid connect post-INPROGRESS
	     *      is flagged by a read or write event, not an except...  
             *
             *      this snippet from the case pl_fd_CONNECTING: of _pl_fd_connect()
	     */
	    result = 0;
	    fdo->state = pl_fd_CONNECTED;
	    result = (fdo->event_connected)( fdo, fdo->el );
	}
    }

    return result;
}

static int _pl_fd_connect(pl_fd *fdo)
{
  int result = -1;
  pl_fd_event_handler callback0 = 0, callback1 = 0;
  int nblock_save = 0;

  switch ( fdo->ossocktype ) {
  case SOCK_STREAM:
    /* SOCK_STREAM needs connect. */
    switch ( fdo->state ) {
    case pl_fd_CONNECTED:
      /* Already connected. */
      result = 0;
      break;

    case pl_fd_CONNECTING:
      /* Connecting: check socket's error status. */
      {
	int so_error = 0;
	socklen_t so_error_len = sizeof(so_error); 
	
	/* Check to see if any errors happened. */
	if ( getsockopt(fdo->fd, SOL_SOCKET, SO_ERROR, &so_error, &so_error_len) ) {
	  fprintf(stderr, "getsockopt(%d): failed: %s\n", fdo->fd, strerror(errno));
	}
#if 0
	fprintf(stderr, "%d SO_ERROR = %d: %s\n", fdo->fd, so_error, strerror(so_error));
#endif

	/* Socket error? */
	if ( so_error ) {
	  result = -1;

	  fdo->oserr = so_error;
	  fdo->oserrstr = 0;

	  /* Socket is unuseable. */
	  pl_fd_close_(fdo);

	  callback0 = fdo->event_connect_error;
	} else {
	  /* Connected. */
	  result = 0;
	  fdo->state = pl_fd_CONNECTED;
	  callback1 = fdo->event_connected;
	}
      }
      break;

    default:

      /*
      ** For async connects, temporarily set non-blocking so connect() will return
      ** immediately.  This is POSIX stuff and probably doesn't belong here.
      */
      if ( fdo->async ) {
	nblock_save = pl_fd_nblock(fdo);
        pl_fd_set_nblock(fdo, 1);
      }

      pl_fd_SYSCALL(result, fdo, connect(fdo->fd, (const struct sockaddr *) &fdo->ossockaddr, sizeof(fdo->ossockaddr)));

      if ( fdo->async ) {
        pl_fd_set_nblock(fdo, nblock_save);
      }

      /* fdo->fd is O_NONBLOCK and connect has started. */
      switch ( result ) {	
      case pl_fd_EINPROGRESS:
	result = 0;
	if ( fdo->state != pl_fd_CONNECTING ) {
	  fdo->state = pl_fd_CONNECTING;
	  callback0 = fdo->event_connecting;
	}
	/* save the exception handler for later restoration */
	fdo->__save_event_except = fdo->event_except;
	fdo->event_except = _pl_fd_cli_connect_inprogress_except;
	break;

      case pl_fd_EALREADY:
	/* connect() already started. */
	result = 0;
	if ( fdo->state != pl_fd_CONNECTING ) {
	  fdo->state = pl_fd_CONNECTING;
	  callback0 = fdo->event_connecting;
	}
	break;

      case 0:
	/* connect() is done. */
	result = 0;
	if ( fdo->state != pl_fd_CONNECTED ) {
	  fdo->state = pl_fd_CONNECTED;
	  callback0 = fdo->event_connecting;
	  callback1 = fdo->event_connected;
	}
	break;

      default:
	/* Some other error occurred. */
	result = -1;

	/* Socket is unuseable. */
	pl_fd_close_(fdo);

	callback0 =  fdo->event_connect_error;
	break;
      }
    }
    break;

  case SOCK_DGRAM:
    /* DGRAM is connectionless. */
    result = 0;
    if ( fdo->state != pl_fd_CONNECTED ) {
      fdo->state = pl_fd_CONNECTED;
      callback0 = fdo->event_connecting;
      callback1 = fdo->event_connected;
    }
    break;

  default:
    /* ??? */
    assert(! "SOCK_STREAM || SOCK_DGRAM");
    break;
  }

  /* Notify. */
  do {
    if ( callback0 && (result = callback0(fdo, fdo->el)) )
      break;
    
    if ( callback1 && (result = callback1(fdo, fdo->el)) )
      break;
  } while(0);

  return result;
}


static int _pl_fd_socket_open(pl_fd *fdo)
{
  int result = -1;

  do {
    const char *x;

    if ( (result = _pl_fd_sockaddr_from_name(fdo)) )
      break;
    
    /* Create the socket. */
    if ( (result = _pl_fd_socket(fdo)) < 0 )
      break;
    
    /* Use a local address on the socket? */
    if ( (x = pl_fd_local_name(fdo)) && *x ) {
      if ( (result = _pl_str_to_sockaddr(x, &(fdo->os_local_sockaddr), 0, 0, 0, 0)) )
	break;

      /* Bind socket to connect using local address. */
      if ( (result = _pl_fd_bind(fdo)) )
	break;
    }

    if ( (result = _pl_fd_connect(fdo)) < 0 )
      break;

    /* Get the local address that was bound by OS implicitly. */
    if ( ! (x && *x) ) {
      socklen_t len = sizeof(fdo->os_local_sockaddr);

      pl_fd_SYSCALL(result, fdo, getsockname(fdo->fd, &(fdo->os_local_sockaddr), &len));
    }
  } while ( 0 );

  return result;
}


int pl_fd_socket(pl_fd *fdo, const char *name)
     /*@ 
       Initializes %fdo% as a client-side socket.
       name is of the form "host:service:protocol".
       service can be a decimal port number or a service name.
       protocol can be a decimal protocol number or protocol name.
       protocol defaults to "tcp" if protocol is null.
       Returns non-zero if error.
     */
{
  assert(fdo);

  /* Save name. */
  fdo->name = realloc(fdo->name, strlen(name) + 1);
  assert(fdo->name);
  strcpy(fdo->name, name);

  /* Set open method. */
  fdo->osopen = _pl_fd_socket_open;

  return 0;
}


#undef pl_fd_SYSCALL


static pl_fd_event_handler_DECL(_pl_fd_cli_event_connected)
     /*@
       Called when a asynchronous connnecting client socket has connected.
     */
{
  int result = 0;

  /* Remove old event handlers. */
  fdo->event_osread = 0;
  fdo->event_oswrite = 0;
  fdo->event_mask &= ~(pl_fd_OSREAD|pl_fd_OSWRITE|pl_fd_OSEXCEPT);
 
  /* Validate connection. */
  result = pl_fd_open(fdo);

  return result;
}

static int _pl_fd_socket_async_open(pl_fd *fdo)
{
  int result;

  do {
    /* Call regular open. */
    result = _pl_fd_socket_open(fdo);

    if (fdo->state == pl_fd_CONNECTING) {
      /* Call _event_connected when connect is done. */
      fdo->event_enable = 1;
      fdo->event_mask |= pl_fd_OSREAD | pl_fd_OSWRITE;
      fdo->event_osread = 
      fdo->event_oswrite = _pl_fd_cli_event_connected;
      break;
    }
    else {
      /* In all cases, reset the osread and write pointers if they
       * are still set to the connecting function, in case we get
       * extraneous data from the user. Normally these won't matter
       */
      if ( fdo->event_osread == _pl_fd_cli_event_connected ) {
        fdo->event_osread = 0;
        fdo->event_mask &= ~ pl_fd_OSREAD;
      }
      if ( fdo->event_oswrite == _pl_fd_cli_event_connected ) {
        fdo->event_oswrite = 0;
        fdo->event_mask &= ~ pl_fd_OSWRITE;
      }
	
      if (fdo->state != pl_fd_CONNECTED) {
      	/* PJC - if it is an error, it should return error */
      	result = 1;
      }
    }

    break;

  } while ( 0 );

  return result;
}


int pl_fd_socket_async(pl_fd *fdo, const char *conn)
     /*@
       Initializes %fdo% to connect asynchronously.
       %fdo%->connecting is called when %fdo% is connecting.
       %fdo%->connected is called when %fdo% is connected.
     */
{
  int result;

  do {
    if ( (result = pl_fd_socket(fdo, conn)) )
      break;
    
    fdo->async = 1;
    fdo->osopen = _pl_fd_socket_async_open;

  } while ( 0 );

  return result;
}


/*******************************************************************************/


char *pl_fd_read_until(pl_fd *fdo, int term)
     /*@ 
       Returns a buffer read until %term%, inclusively, EOF or error.
       Returns 0 if already at EOF.
       Must call free() with result.
      */
{
  int i = 0;
  char *buf = malloc(i + 2);
  int result;

  assert(buf);

  do {
    result = pl_fd_read(fdo, buf + i, 1);
  } while ( (result > 0) && (buf[i] != term) && (++ i) && (buf = realloc(buf, i + 2)));
  buf[i + 1] = 0;

  if ( i == 0 && result <= 0 ) {
    free(buf);
    return 0;
  }
  
  return buf;
}


/**********************************************************************/
/* EOF. */
