/* Author: kurt.stephens@parlano.com 2000/10/09 */
/*   parlano eventloop library -- pl_rr/buf.c
 *   Copyright (C) 2001 Parlano, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "reqres.h"

#include <stdlib.h> /* malloc(), free(), realloc() */
#include <stdio.h>
#include <assert.h>

/*************************************************************************/

int pl_rr_buf_resize(pl_rr_buf *rrb, size_t bufsize)
     /*@ Resizes %rrb% to %bufsize%. */
{
  char *buf;

  assert(rrb);

  buf = realloc(rrb->buf, bufsize + 1);
  assert(buf);

  assert(rrb->readp >= rrb->buf);
  rrb->readp = buf + (rrb->readp - rrb->buf);
  assert(rrb->writep >= rrb->buf);
  rrb->writep = buf + (rrb->writep - rrb->buf);

  rrb->buf = buf;
  rrb->end = rrb->buf + bufsize;
  rrb->bufsize = bufsize;
  rrb->buf[bufsize] = 0;

  assert(rrb->readp <= rrb->end);
  assert(rrb->writep <= rrb->end);

  return 0;
}


int pl_rr_buf_init(pl_rr_buf *rrb, pl_rr_id part, const char *buf, size_t len )
/*@ 
  Initialize a buffer with a buffer %buf% of %len% bytes.
  %buf% can be null.
*/
{
  assert(rrb);

  /* Initialize for first resize() */
  rrb->buf = 0;
  rrb->readp = 0;
  rrb->writep = 0;
  rrb->state = pl_rr_buf_START;
  rrb->part = part;

  /* Resize to len. */
  pl_rr_buf_resize(rrb, len);

  /* Append buf. */
  if ( buf ) {
    memcpy(rrb->writep, buf, len);
    rrb->writep += len;
  }

  return 0;
}


int pl_rr_buf_destroy(pl_rr_buf *rrb)
     /*@ Destroys any data inside %rrb%. */
{
  assert(rrb);

  free(rrb->buf);
  rrb->buf = 0;

  return 0;
}



int _pl_rr_buf_reset(pl_rr_buf *rrb)
     /*@ Resets %rrb% for reading and writing. */
{
  assert(rrb);

  rrb->readp = rrb->writep = rrb->buf;
  rrb->state = pl_rr_buf_START;

  return 0;
}



/********************************************************************/


int pl_rr_buf_close(pl_rr_buf *rrb)
/*@
  If the buffer is:
    in START state, mark it as COMPLETE.
    in IO_PENDING state, mark it as IO_COMPLETE.
 */
{
  assert(rrb);

  switch ( rrb->state ) {
  case pl_rr_buf_START:
    rrb->state = pl_rr_buf_COMPLETE;
    break;

  case pl_rr_buf_IO_PENDING:
    rrb->state = pl_rr_buf_IO_COMPLETE;
    break;

 default:
    assert(! "valid state");
    abort();
    break;
  }

  return 0;
}


pl_rr_buf_state pl_rr_buf_get_state(pl_rr_buf *rrb)
/*@
  Returns the state of the buffer.
 */
{
  assert(rrb);

  return rrb->state;
}



/********************************************************************/

char *pl_rr_buf_write_ptr(pl_rr_buf *rrb)
     /*@ Returns the start of the writable location in %rrb%. */
{
  assert(rrb);
  assert(rrb->writep);

  return rrb->writep;
}


size_t pl_rr_buf_write_len(pl_rr_buf *rrb)
     /*@ Returns the space left to write at the writeable location in %rrb%. */
{
  assert(rrb);

  return rrb->end - rrb->writep;
}


int pl_rr_buf_write_count(pl_rr_buf *rrb, size_t count)
     /*@ 
       Increments the write location as if %count% bytes were written to %rrb%.  
       Does not grow the buffer. 
       */
{
  assert(rrb);
  assert(pl_rr_buf_write_len(rrb) >= count);

  rrb->writep += count;
  *(rrb->writep) = 0;

  return 0;
}


int pl_rr_buf_write(pl_rr_buf *rrb, const char *buf, size_t len )
     /*@ Writes to the buffer, growing it if necessary. */
{
  size_t left = rrb->end - rrb->writep;

  assert(rrb);
  assert(rrb->state == pl_rr_buf_START);

  /* Zero length writes are okay. */
  if ( ! len )
    return 0;

  /* Resize if necessary. */
  if ( len > left ) {
    pl_rr_buf_resize(rrb, rrb->bufsize + (len - left));
  }

  /* Make sure room is available. */
  assert(rrb->writep + len <= rrb->end);

  /* Copy bytes from buf into write pointer. */
  assert(buf);
  memcpy(rrb->writep, buf, len);
  rrb->writep += len;
  *(rrb->writep) = 0;
  
  return 0;
}


/********************************************************************/


char *pl_rr_buf_read_ptr(pl_rr_buf *rrb)
     /*@ Returns the read location from %rrb%. */
{
  assert(rrb);
  assert(rrb->readp);

  return rrb->readp;
}


size_t pl_rr_buf_read_len(pl_rr_buf *rrb)
     /*@ Return the number of bytes available to read from %rrb%. */
{
  assert(rrb);
  assert(rrb->writep > rrb->readp);

  return rrb->writep - rrb->readp;
}


int pl_rr_buf_read_count(pl_rr_buf *rrb, size_t count)
     /*@ Advance the read location as if %count% bytes were read from %rrb%. */
{
  assert(rrb);
  assert(pl_rr_buf_read_len(rrb) >= count);

  rrb->readp += count;

  return 0;
}


/******************************************************************/


int pl_rr_buf_dump(pl_rr_buf *rrb, const char *prefix, FILE *fp)
{
  char buf[256], *s;

  assert(rrb);
  assert(fp);

  /* Show buffer and size. */
  strcpy(buf, rrb->buf);
  for ( s = buf; *s; ++ s ) {
    if ( ! (32 <= *s && *s < 127) ) {
      *s = '.';
    }
  }

  fprintf(fp, "%s%d '%s' %lu\n", 
	  prefix,
	  (int) rrb->state, 
	  buf,
	  (unsigned long) rrb->bufsize);

  /* Show read and write pointers. */
  assert(sizeof(buf) >= rrb->bufsize);
  memset(buf, ' ', sizeof(buf));
  buf[rrb->bufsize + 1] = 0;
  
  buf[rrb->writep - rrb->buf] = 'W';
  if ( rrb->writep == rrb->readp ) {
    buf[rrb->readp - rrb->buf] = '+';
  } else {
    buf[rrb->readp - rrb->buf] = 'R';
  }

  fprintf(fp, "%s%s |%s|\n", prefix, " ", buf); 

  return 0;
}


