/* Author: kurt.stephens@parlano.com 2000/10/09 */
/*   parlano eventloop library -- pl_rr/endpoint.c
 *   Copyright (C) 2001 Parlano, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "hdr.h"
#include "reqres.h"

#include <stdlib.h> /* malloc(), free(), realloc() */
#include <stdio.h>
#include <assert.h>

/*************************************************************************/

#ifndef DEBUG
#define DEBUG ((ep)->debug)
#endif

int pl_rr_endpoint_init(pl_rr_endpoint *ep, pl_rr_q *queue, pl_rr_handler reqh, void *reqh_data)
{
  assert(ep);
  assert(queue);

  memset(ep, 0, sizeof(*ep));

  ep->queue = queue;
  ep->reqh = reqh;
  ep->reqh_data = reqh_data;

  return 0;
}


int pl_rr_endpoint_destroy(pl_rr_endpoint *ep)
{
  assert(ep);

  memset(ep, 0, sizeof(*ep));

  return 0;
}


/******************************************************************/

int _pl_rr_endpoint_sendable(pl_rr_endpoint *ep, pl_rr *rr)
{
  int result;

  assert(ep);
  assert(ep->sendable);

  result = ep->sendable(ep, rr);

  return result;
}

int pl_rr_endpoint_clear_pending(pl_rr_endpoint *ep)
     /* Drop all pending reads and writes */
{
  assert(ep);

  if ( DEBUG ) {
    fprintf(stderr, "pl_rr_endpoint_clear(%p)\n", ep);
  }

  /* Put back in start state. */
  ep->state[0] = ep->state[0] = pl_rr_ep_START;
  ep->rr[0] = ep->rr[1] = 0;
  ep->rrb[0] = ep->rrb[1] = 0;

  /* Destroy all outstanding requests */
  pl_rr_q_destroy(ep->queue);

  return 0;
}

int pl_rr_endpoint_reschedule(pl_rr_endpoint *ep)
     /*@
       Force any current reads or writes to be dropped and rescheduled.
       Called after IO error (connection dropped).
      */
{
  assert(ep);

  if ( DEBUG ) {
    fprintf(stderr, "pl_rr_endpoint_reschedule(%p)\n", ep);
  }

  /* Put back in start state. */
  ep->state[0] = ep->state[0] = pl_rr_ep_START;
  ep->rr[0] = ep->rr[1] = 0;
  ep->rrb[0] = ep->rrb[1] = 0;

  /* Reset all request for endpoint. */
  _pl_rr_q_reschedule(ep->queue, ep);

  return 0;
}



/******************************************************************/
/* Buffer-level I/O */

static
int _pl_rr_endpoint_send(pl_rr_endpoint *ep, const void *buf, size_t len)
{
  int result;

  assert(ep);

  result = ep->send(ep, buf, len, ep->send_data);

  /* EOF on write is an error. */
  if ( result == 0 ) {
    assert(ep->eos);
    assert(ep->ioerror);
    (void) (ep->ioerror(ep) || ep->eos(ep));
    return -1;
  } 
  /* I/O error */
  else if ( result == -1 ) {
    assert(ep->ioerror);
    ep->ioerror(ep);
    return -1;
  }

  return result;
}


int pl_rr_endpoint_send(pl_rr_endpoint *ep)
     /*@
       Sends any outgoing data, if available.
     */
{
  int dir = pl_rr_OUTGOING;
 
  pl_rr_header *h;
  pl_rr *rr;
  pl_rr_buf *rrb;

  assert(ep);

  h = &(ep->hdr[dir]);
  rr = ep->rr[dir];
  rrb = ep->rrb[dir];

  /* Write any OUTGOING messages. */

  /* This code will not work with UDP because each send() causes a new datagram to be sent.
   * Remedy:
   *  1. buffer output here till header and body are both ready.
   *  2. buffer output in fdo.
   *  3. use sendv().
   *
   * a. Buffered IO in fdo might make more sense, since other UDP apps will be able to
   * take advantage of it, assuming they know when to call pl_fd_flush().
   *
   * -- kurt.stephens@parlano.com 2000/10/09
   */

  switch ( ep->state[dir] ) {
  case pl_rr_ep_START:
    {
      /* Look for any queued request/response output for this endpoint. */
      rr = _pl_rr_q_next_sendable(ep->queue, ep);

      assert(ep->rr[dir] == 0);
      assert(ep->rrb[dir] == 0);

      if ( rr ) {
	/* Fill in the header. */
	assert(0 <= rr->dir && rr->dir <= 1);
	rrb = &(rr->buf[rr->dir]);
	assert(rrb->state == pl_rr_buf_COMPLETE);
	rr->state = rr->dir == pl_rr_OUTGOING ? pl_rr_REQUEST_IO : pl_rr_RESPONSE_IO;
	rrb->state = pl_rr_buf_IO_PENDING;

	h->proto = rr->proto;
	h->id = rr->id;
	h->op = rr->dir;
	h->part = rrb->part;
	h->len = pl_rr_buf_read_len(rrb);

	/* Encode header. */
	pl_rr_header_encode(h);
	
	/* Begin writing the reqres's header. */
	ep->state[dir] = pl_rr_ep_HEADER;
	ep->rr[dir] = rr;
	ep->rrb[dir] = rrb;

	/* FALL THROUGH */
      } else {
	/* Nothing is sendable. */
	_pl_rr_endpoint_sendable(ep, 0);
	break;
      }
    }
    /* FALL THROUGH */

  case pl_rr_ep_HEADER:
    if ( h->rp < h->end ) {
      int result;

      result = _pl_rr_endpoint_send(ep, h->rp, h->end - h->rp);
      
      /* ERROR? */
      if ( result == -1 ) {
	/* Retransmit. */
	h->rp = h->buf;
	return result;
      }
      /* Increment read position. */
      else if ( result > 0 ) {
	h->rp += result;
      }
    }

    /* Has the entire header been sent? */
    if ( h->rp >= h->end ) {
      ep->state[dir] = pl_rr_ep_BODY;

      /* FALL THROUGH */
    } else {
      break;
    }
    /* FALL THROUGH */

  case pl_rr_ep_BODY:
    {
      size_t left;

      assert(ep->rr[dir]);
      assert(0 <= h->op && h->op <= 1);

      left = pl_rr_buf_read_len(rrb);
      if ( left ) {
	char *buf = pl_rr_buf_read_ptr(rrb);
	int result;

	result = _pl_rr_endpoint_send(ep, buf, left);

	/* ERROR? */
	if ( result == -1 ) {
	  /* Retransmit. */
	  h->rp = h->buf;
	  return result;
	}
	else if ( result >= 0 ) {
	  pl_rr_buf_read_count(rrb, result);
	  left -= result;
	}

	/* Was the entire message sent? */
	if ( left == 0 ) {
	  /* Close the request. */
	  _pl_rr_close(rr, h->op);

	  /* Go back to START state. */
	  ep->state[dir] = pl_rr_START;
	  ep->rr[dir] = 0;
	  ep->rrb[dir] = 0;
	}
      }
    }
    break;

  default:
    abort();
    break;
  }

  return 0;
}



static
int _pl_rr_endpoint_recv(pl_rr_endpoint *ep, void *buf, size_t len)
{
  int result;

  assert(ep);

  result = ep->recv(ep, buf, len, ep->recv_data);

  /* EOF on read is an error, in caller's context. */
  if ( result == 0 ) {
    assert(ep->ioerror);
    ep->ioerror(ep);
    return -1;
  } 
  /* I/O error */
  else if ( result == -1 ) {
    assert(ep->ioerror);
    ep->ioerror(ep);
    return -1;
  }

  return result;
}


int pl_rr_endpoint_recv(pl_rr_endpoint *ep)
{
  int dir = pl_rr_INCOMING;
  pl_rr_header *h;
  pl_rr *rr;
  pl_rr_buf *rrb;
  int result;
  int rtn = 0;

  assert(ep);

  h = &ep->hdr[dir];
  rr = ep->rr[dir];
  rrb = ep->rrb[dir];

  /* A request or a response is readable. */
  
  switch ( ep->state[dir] ) {
  case pl_rr_ep_START:
    {
      pl_rr_header_init(h);
      ep->state[dir] = pl_rr_ep_HEADER;
    }
    /* FALL THROUGH */

  case pl_rr_ep_HEADER:
    assert(h);
    assert(h->end > h->rp);

    /* Read some bytes of the header. */
    result = ep->recv(ep, h->rp, h->end - h->rp, ep->recv_data);
      
    /* EOS? */
    if ( result == 0 ) {
      /* EOS at beginning of HEADER is ok, but notify for cleanup. */
      assert(ep->eos);
      rtn = ep->eos(ep);
      break;
    }
    /* Read error occurred. */
    else if ( result == -1 ) {
      /* Retry read. */
      /* Return to START state. */
      ep->state[dir] = pl_rr_ep_START;
      ep->rr[dir] = 0;
      ep->rrb[dir] = 0;
      rtn = ep->ioerror(ep);
      break;
    }
    else if ( result > 0 ) {
      /* Increment. */
      h->rp += result;
    }
    
    assert(ep->rr[dir] == 0);
    assert(ep->rrb[dir] == 0);

    /* Did we read the header? */
    if ( h->rp >= h->end ) {
      /* Decode the header. */
      pl_rr_header_decode(h);
	
      /* Switch to reading a part BODY. */
      ep->state[dir] = pl_rr_ep_BODY;
	
      /* Find or create a request object to read the BODY. */
      rr = ep->rr[dir] 
	= _pl_rr_q_begin_read(ep->queue, 
				  h->op, 
				  h->id, 
				  h->part, 
				  h->len,
				  ep
				 );

      /* Get the buffer for the request object. */
      assert(rr);
      rrb = ep->rrb[dir] = &(rr->buf[h->op]);
      _pl_rr_buf_reset(rrb);
      rrb->part = h->part;
      rr->state = rr->dir == pl_rr_INCOMING ? pl_rr_REQUEST_IO : pl_rr_RESPONSE_IO;
      rrb->state = pl_rr_buf_IO_PENDING;
      /* FALL THROUGH */
    } else {
      break;
    }
    /* FALL THROUGH */

  case pl_rr_ep_BODY:
    {
      size_t left;
      
      assert(rrb);
      
      /* How many bytes left to read? */
      left = pl_rr_buf_write_len(rrb);
      if ( left ) {
	/* Get a write ptr to read into. */
	char *buf = pl_rr_buf_write_ptr(rrb);
	int result;
	
	/* Read some bytes of the body. */
	result = _pl_rr_endpoint_recv(ep, buf, left);
	
	/* Error out. */
	if ( result == -1 ) {
	  /* Return to START state. */
	  ep->state[dir] = pl_rr_ep_START;
	  ep->rr[dir] = 0;
	  ep->rrb[dir] = 0;
	  return 1;
	}
	else if ( result > 0 ) {
	  /* "result" bytes was actually read. */
	  pl_rr_buf_write_count(rrb, result);
	  left -= result;
	}
      }
      
      /* If this is the end of the part, */
      if ( left == 0 ) {
	/* Mark the buffer as complete. */
	_pl_rr_close(rr, h->op);
	
	/* Return to START state. */
	ep->state[dir] = pl_rr_ep_START;
	ep->rr[dir] = 0;
	ep->rrb[dir] = 0;
      }
    }
    break;
    
  default:
    abort();
    break;
  }

  return rtn;
}

