/* Author: kurt.stephens@parlano.com 2000/09/26 */
/*   parlano eventloop library -- pl_rr/fdobj_reqres.c
 *   Copyright (C) 2001 Parlano, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "fdobj_reqres.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* strerror() */
#include <assert.h>

#ifndef DEBUG
#define DEBUG 0
#endif


pl_rr_fdo *pl_fd_rr_fdo(pl_fd *fdo)
     /*@ Get the pl_rr_fdo object from a pl_fd object. */
{
  assert(fdo);

  return (pl_rr_fdo*) fdo->opaque[0];
}


pl_rr_fdo *pl_rr_endpoint_rr_fdo(pl_rr_endpoint *ep)
     /*@ Get the pl_rr_fdo object from a pl_rr_endpoint object. */
{
  assert(ep);

  return (pl_rr_fdo*) ep->opaque[0];
}


const char *_n(pl_fd *fdo)
{
  return ((pl_rr_fdo*) (fdo->opaque[0]))->name;
}

/******************************************************************/
/* Event handlers. */

static pl_fd_event_handler_DECL(_event_write)
     /*@ Called when write is possible on %fdo%. */
{
  pl_rr_fdo *fdrr = pl_fd_rr_fdo(fdo);

  if ( DEBUG ) {
    fprintf(stderr, "%s: _event_write(%p, %p)\n", _n(fdo), fdo, el);
  }

  assert(fdrr);

  /* Have the req/res queue handle the write. */
  return pl_rr_endpoint_send(fdrr->ep);
}


static pl_fd_event_handler_DECL(_event_read)
     /*@ Called when read is possible on %fdo%. */
{
  pl_rr_fdo *fdrr = pl_fd_rr_fdo(fdo);

  if ( DEBUG ) {
    fprintf(stderr, "%s: _pl_rr_fd_event_read(%p, %p)\n", _n(fdo), fdo, el);
  }

  assert(fdrr);

  /* Have the req/res queue handle the read. */
  return pl_rr_endpoint_recv(fdrr->ep);
}


static pl_fd_event_handler_DECL(_event_except)
     /*@ Called when an exceptional I/O condition is on %fdo%. */
{
  pl_rr_fdo *fdrr = pl_fd_rr_fdo(fdo);

  if ( DEBUG ) {
    fprintf(stderr, "%s: _pl_rr_fd_event_except(%p, %p)\n", _n(fdo), fdo, el);
  }

  assert(fdrr);

  /* IMPLEMENT. */
  abort();

  return 1; /* Don't continue with any other event handlers. */
}


/******************************************************************/
/* pl_rr_endpoint to pl_fdrr glue. */


static int _sendable(pl_rr_endpoint *ep, pl_rr *rr)
     /*@ Called when %ep% has a request or response that can be written. */
{
  pl_rr_fdo *fdrr = pl_rr_endpoint_rr_fdo(ep);
  pl_fd *fdo;

  assert(fdrr);

  fdo = fdrr->fdo;
  
  if ( DEBUG ) {
    fprintf(stderr, "%s: _sendable(%p, %p)\n", _n(fdo), ep, rr);
  }

  if ( rr ) {
    /* Force writing on next eventloop. */
    fdo->event_mask |= pl_fd_WRITE;
  } else {
    /* Disable writing on next eventloop. */
    fdo->event_mask &= ~ pl_fd_WRITE;
  }

  return 0;
}


static int _send(pl_rr_endpoint *ep, const void *buf, size_t size, void *data)
     /*@ Called from %ep% when writing request/response data. */
{
  pl_rr_fdo *fdrr = pl_rr_endpoint_rr_fdo(ep);
  pl_fd *fdo = fdrr->fdo;
  int result;

  if ( DEBUG ) {
    fprintf(stderr, "%s: _send(%p, %p, %lu)\n", _n(fdo), ep, buf, (unsigned long) size);
  }

  assert(fdrr);
  assert(fdo);

  result = pl_fd_write(fdo, buf, size);

  return result;
}
 

static int _recv(pl_rr_endpoint *ep, void *buf, size_t size, void *data)
     /*@ Called from %ep% when writing request/response data. */
{
  pl_rr_fdo *fdrr = pl_rr_endpoint_rr_fdo(ep);
  pl_fd *fdo = fdrr->fdo;
  int result;

  if ( DEBUG ) {
    fprintf(stderr, "%s: _recv(%p, %p, %lu)\n", _n(fdo), ep, buf, (unsigned long) size);
  }

  assert(fdrr);
  assert(fdo);

  result = pl_fd_read(fdo, buf, size);

  return result;
}


static int _terminate_or_reopen(pl_rr_endpoint *ep, int reopen)
     /*@ 
       If %ep% is an accepted client, 
       destroy the connection.
       If %ep% is a connecing client,
       schedule a reconnect if %reopen% is non-zero,
       otherwise close the connection.
     */
{
  pl_rr_fdo *fdrr = pl_rr_endpoint_rr_fdo(ep);
  pl_fd *fdo = fdrr->fdo;

  if ( fdrr->srv ) {
    /* Server-side accepted connection should drop connection, 
       because client did clean EOS.
     */
    pl_fd_destroy(fdo);
  } else {
    if ( reopen ) {
#if 0
      /* Reschedule any requests for resend after reconnect. */
      /* 
	 User callbacks should decide if rr's should be 
	 rescheduled or aborted.
	 -- kurt.stephens@parlano.com 2001/01/19
      */
      pl_rr_endpoint_reschedule(ep);
#endif

      /* Schedule client reconnection. */
      pl_fd_reopen(fdo);
    } else {
      /* Close client connection. */
      pl_fd_close(fdo);
    }
  }

  return 0;
}


static int _ioerror(pl_rr_endpoint *ep)
     /*@ Called when an I/O error occurs on fdrr->fdo. */
{
  pl_rr_fdo *fdrr = pl_rr_endpoint_rr_fdo(ep);
  pl_fd *fdo = fdrr->fdo;
  int result = 0;

  if ( fdrr->debug ) {
    fprintf(stderr, "%s: IO ERROR: %s\n", fdo->name, pl_fd_get_oserrstr(fdo));
  }

  /* Event handler. */
  if ( fdrr->event_ioerror ) {
    result = fdrr->event_ioerror(fdrr);
  }

  /* It was an I/O error, now what? */
  _terminate_or_reopen(ep, result ? 0 : 1);

  return 1;
}


static int _eos(pl_rr_endpoint *ep)
     /*@ Called when a EOF occurs on fdrr->fdo. */
{
  pl_rr_fdo *fdrr = pl_rr_endpoint_rr_fdo(ep);
  pl_fd *fdo = fdrr->fdo;
  int result = 0;

  if ( fdrr->debug ) {
    fprintf(stderr, "%s: EOS\n", fdo->name);
  }

  /* Event handler. */
  if ( fdrr->event_eos ) {
    result = fdrr->event_eos(fdrr);
  }

  _terminate_or_reopen(ep, result ? 0 : 1);

  return 1;
}


/******************************************************************/
/* pl_rr_endpoint I/O hooks. */


static int _fdo_destroy(pl_fd *fdo)
     /*@ Called when pl_fd_destroy() is called on %fdo% as initialized by %_pl_rr_fdo_init%(). */
{
  pl_rr_fdo *rrfd = pl_fd_rr_fdo(fdo);

  if ( rrfd ) {
    pl_rr_fdo_destroy(rrfd);
  }
  
  return 0;
}


static int _pl_rr_fdo_init
(
 pl_rr_fdo *fdrr,
 pl_rr_q *queue,
 pl_rr_handler reqh, 
 void *reqh_data,
 pl_el *el,
 pl_fd *fdo
 )
{
  int result;

  assert(fdrr);
  /* assert(reqh); */
  assert(el);

  memset(fdrr, 0, sizeof(*fdrr));

  /* Use space in fdrr for fdo. */
  if ( ! fdo ) {
    fdo = &fdrr->_fdo;

    /* Initialize a server connection. */
    if ( (result = pl_fd_init(fdo)) )
      return -1;
  }

  fdrr->fdo = fdo;

  /* Destroy fdrr if fdo is destroyed. */
  fdo->destroy = _fdo_destroy;

  /* Associate fdobj with fdrr. */
  fdo->opaque[0] = fdrr;

  /* Associate queue with fdrr. */
  fdrr->queue = queue;

  /* Initialize request/response endpoint. */
  fdrr->ep = &fdrr->_ep;
  if ( pl_rr_endpoint_init(fdrr->ep, fdrr->queue, reqh, reqh_data) )
    return -1;

  /* Set up req/res queue glue. */
  fdrr->ep->opaque[0]= fdrr;
  fdrr->ep->send = _send;
  fdrr->ep->recv = _recv;
  fdrr->ep->sendable = _sendable;
  fdrr->ep->eos = _eos;
  fdrr->ep->ioerror = _ioerror;

  /* Set up events. */
  fdo->event_enable = 0;
  pl_el_add(el, fdo);
  fdrr->el = el;

  return 0;
}


int pl_rr_fdo_destroy(pl_rr_fdo *fdrr)
     /*@ Destroys all data assocated with fdrr. */
{
  assert(fdrr);

  /* Event handler. */
  if ( fdrr->event_destroy ) {
    pl_rr_fdo_event_handler x = fdrr->event_destroy;
    fdrr->event_destroy = 0;
    x(fdrr);
  }

  /* Remove associations. */
  fdrr->fdo->opaque[0] = 0;

  /* Destroy fdo. */
  if ( fdrr->fdo ) {
    /* Avoid recursion. */
    pl_fd *fdo = fdrr->fdo;
    fdrr->fdo = 0;

    /* Destroy fdo. */
    pl_fd_destroy(fdo);
  }
  
  /* Destroy endpoint. */
  pl_rr_endpoint_destroy(fdrr->ep);

  /* Dealloc. */
  if ( fdrr->dealloc ) {
    fdrr->dealloc(fdrr);
  }

  return 0;
}


/********************************************************************/

static pl_fd_event_handler_DECL(_event_cli_accepting)
     /*@
       Called when accepting fdo.
     */
{
  int result = 0;
  pl_rr_fdo *fdrr = pl_fd_rr_fdo(fdo->accept_from);
  /* The server's glue object. */

  pl_rr_fdo *fdrrc;
  /* The client's glue object. */

  if ( fdrr->debug ) {
    fprintf(stderr, "%s: %s: accepting\n", _n(fdo), fdo->name);
  }

  do {
    /* Allocate new client pl_rr_fdo. */
    fdrrc = malloc(sizeof(*fdrrc));
    if ( ! fdrrc ) {
      fprintf(stderr, "%s: %s: out of memory\n", _n(fdo), fdo->name);
      pl_fd_destroy(fdo);
      result = -1;
      break;
    }

    /* Use server's queue, request hander, and event loop. */
    _pl_rr_fdo_init(fdrrc, fdrr->queue, fdrr->ep->reqh, fdrr->ep->reqh_data, fdrr->el, fdo);

    fdrrc->srv = fdrr;
    fdrrc->dealloc = pl_delayed_free;
    fdrrc->name = fdrrc->srv->name;
    fdrrc->debug = fdrr->debug;

    fdrrc->event_accepting = fdrr->event_accepting;
    fdrrc->event_accepted = fdrr->event_accepted;
    fdrrc->event_accept_error = fdrr->event_accept_error;

    /* Event handler. */
    if ( fdrrc->event_accepting ) {
      result = fdrrc->event_accepting(fdrrc);
      if ( result ) {
	pl_rr_fdo_destroy(fdrrc);
	break;
      }
    }

  } while(0);

  return result;
}


static pl_fd_event_handler_DECL(_event_cli_accepted)
     /*@
       Called when fdo is actually in ACCEPTED state.
     */
{
  int result = 0;
  pl_rr_fdo *fdrr = pl_fd_rr_fdo(fdo);
  /* The client's glue object. */

  if ( fdrr->debug ) {
    fprintf(stderr, "%s: %s: accepted\n", _n(fdo), fdo->name);
  }

  do {
    /* Enable events. */
    fdo->event_enable = 1;
    fdo->event_read = _event_read;
    fdo->event_write = _event_write;
    fdo->event_except = _event_except;
    fdo->event_mask = pl_fd_READ | pl_fd_EXCEPT;

    /* Event handler. */
    if ( fdrr->event_accepted ) {
      result = fdrr->event_accepted(fdrr);

      /* Event handler wants destruction. */
      if ( ! result ) {
	pl_rr_fdo_destroy(fdrr);
	break;
      }
    }

  } while ( 0 );

  return result;
}


static pl_fd_event_handler_DECL(_event_cli_accept_error)
     /*@
       Called when accept fails.
     */
{
  int result = 0;
  pl_rr_fdo *fdrr = pl_fd_rr_fdo(fdo);
  /* The client's glue object. */

  if ( fdrr->debug ) { 
    fprintf(stderr, "%s: %s: accept failed: %s\n", _n(fdo), fdo->name, pl_fd_get_oserrstr(fdo));
  }

  /* Event handler. */
  if ( fdrr->event_accept_error ) {
    result = fdrr->event_accept_error(fdrr);
  }

  pl_fd_destroy(fdo);

  return result;
}




int pl_rr_fdo_init_srv
(
 pl_rr_fdo *fdrr, 
 pl_rr_q *queue,
 pl_rr_handler reqh, 
 void *reqh_data, 
 const char *srv_conn, 
 pl_el *el,
 pl_fd *cli_proto
 )
{
  int result = -1;
  pl_fd *fdo;

  do {
    assert(reqh);

    if ( (result = _pl_rr_fdo_init(fdrr, queue, reqh, reqh_data, el, 0)) )
      break;

    fdo = fdrr->fdo;
    fdrr->srv = fdrr;
    
#if 0
    fdo->debug = 1;
#endif
    cli_proto->opaque[0] = fdrr;
    cli_proto->event_accepting = _event_cli_accepting;
    cli_proto->event_accepted = _event_cli_accepted;
    cli_proto->event_accept_error = _event_cli_accept_error;

    /* Initialize as a server-side bound socket. */
    if ( (result = pl_fd_socket_server_async(fdo, srv_conn, cli_proto)) )
      break;

    /* Set non-blocking for non-blocking accepts(). */
    if ( (result = pl_fd_set_nblock(fdo, 1)) )
      break;
    
    /* Do bind. */
    if ( (result = pl_fd_open(fdo)) )
      break;

  } while ( 0 );

  if ( result ) {
    if ( fdo->debug ) {
      fprintf(stderr, "%s: %s\n", fdo->name, pl_fd_get_oserrstr(fdo));
    }
  }

  return result;
}



/*************************************************************/
/* Client-side connection. */

static pl_fd_event_handler_DECL(_event_cli_connecting)
     /*@
       Called when client socket is connecting.
     */
{
  pl_rr_fdo *fdrr = pl_fd_rr_fdo(fdo);

  int result = 0;

  if ( fdrr->debug ) {
    fprintf(stderr, "%s: connecting\n", fdo->name);
  }

  /* Event handler. */
  if ( fdrr->event_connecting ) {
    result = fdrr->event_connecting(fdrr);
  }

  return result;
}


static pl_fd_event_handler_DECL(_event_cli_connected)
     /*@
       Called when a connnecting client socket has connected.
     */
{
  pl_rr_fdo *fdrr = pl_fd_rr_fdo(fdo);
  
  int result = 0;

  if ( fdrr->debug ) {
    fprintf(stderr, "%s: connected\n", fdo->name);
  }

  /* Enable read events. */
  fdo->event_enable = 1;
  fdo->event_mask = pl_fd_READ | pl_fd_EXCEPT;
  fdo->event_read = _event_read;
  fdo->event_write = _event_write;
  fdo->event_except = _event_except;

  /* There maybe writes already queued. */
  fdo->event_mask |= pl_fd_WRITE;
  pl_rr_endpoint_reschedule(fdrr->ep);
  pl_rr_endpoint_send(fdrr->ep);

  /* Event handler. */
  if ( fdrr->event_connect ) {
    result = fdrr->event_connect(fdrr);
  }

  return result;
}


static pl_fd_event_handler_DECL(_event_cli_connect_error)
     /*@
       Called when a connnecting client socket has connected.
     */
{
  pl_rr_fdo *rrfd = pl_fd_rr_fdo(fdo);
  
  int result = 0;

  if ( rrfd->debug ) {
    fprintf(stderr, "%s: connect error: %s\n", fdo->name, pl_fd_get_oserrstr(fdo));
  }

  /* Event handler. */
  if ( rrfd->event_connect_error ) {
    result = rrfd->event_connect_error(rrfd);
  }

  return result;
}


static pl_fd_event_handler_DECL(_event_retry_open)
     /*@
       Called when a connnecting client socket has connected.
     */
{
  pl_rr_fdo *rrfd = pl_fd_rr_fdo(fdo);

  int result = 0;

  if ( rrfd->debug ) {
    fprintf(stderr, "%s: retrying: %s\n", fdo->name, pl_fd_get_oserrstr(fdo));
  }

  if ( rrfd->event_connect_retry ) {
    result = rrfd->event_connect_retry(rrfd);
  }

  return result;
}


static pl_fd_event_handler_DECL(_event_retry_open_error)
{
  pl_rr_fdo *rrfd = pl_fd_rr_fdo(fdo);

  int result = 0;

  if ( rrfd->debug ) {
    fprintf(stderr, "%s: giving up: %s\n", fdo->name, pl_fd_get_oserrstr(fdo));
  }

  if ( rrfd->event_connect_retry_error ) {
    result = rrfd->event_connect_retry_error(rrfd);
  }

  return result;
}


int pl_rr_fdo_init_cli
(
 pl_rr_fdo *fdrr, 
 pl_rr_q *queue,
 pl_rr_handler reqh, 
 void *reqh_data,
 const char *cli_conn, 
 pl_el *el,
 int reconnect_delay 
 )
{
  int result = -1;
  pl_fd *fdo;

  do {
    /* Create a req/res queue using the client connection. */
    if ( (result = _pl_rr_fdo_init(fdrr, queue, reqh, reqh_data, el, 0)) )
      break;

    fdo = fdrr->fdo;

    /* Non-blocking. */
    if ( (result = pl_fd_set_nblock(fdo, 1)) )
      break;
    
    /* Create socket. */
    fdo->retry_open_max = reconnect_delay ? -1 : 0;
    fdo->retry_open_delay = reconnect_delay;
    fdo->event_retry_open = _event_retry_open;
    fdo->event_retry_open_error = _event_retry_open_error;

    fdo->event_connecting = _event_cli_connecting;
    fdo->event_connected = _event_cli_connected;
    fdo->event_connect_error = _event_cli_connect_error;

    if ( (result = pl_fd_socket_async(fdo, cli_conn)) )
      break;

    /* Attempt connect now. */
    if ( (result = pl_fd_open(fdo)) )
      break;

  } while ( 0 );

  return result;  
}

