/* Author: kurt.stephens@parlano.com 2000/10/09 */
/*   parlano eventloop library -- pl_rr/queue.c
 *   Copyright (C) 2001 Parlano, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "hdr.h"
#include "reqres.h"

#include "pl_el/dlinklist.h"
#include "pl_el/pl_el.h" /* pl_delayed_free() */

#include <stdlib.h> /* malloc(), free(), realloc() */
#include <stdio.h>
#include <assert.h>

/*************************************************************************/


#ifndef PL_RR_Q_DEBUG
#define PL_RR_Q_DEBUG ((rrq)->debug)
#endif


int pl_rr_q_init(pl_rr_q *rrq)
     /*@
       Initializes %rrq% request/response queue.
     */
{
  assert(rrq);

  memset(rrq, 0, sizeof(*rrq));

  pl_dll_INIT(&(rrq->head));

  return 0;
}


int pl_rr_q_destroy(pl_rr_q *rrq)
     /*@
       Destroys the %rrq% object.
     */
{
  pl_rr *rr;

  assert(rrq);

  if ( PL_RR_Q_DEBUG ) {
    if ( rrq->count ) {
      fprintf(stderr, "pl_rr_q_destroy(%p): still has %d items\n", rrq, rrq->count);
      /*
	pl_rr_q_dump(rrq, stderr);
      */
    }
  }

  pl_rr_q_LOOP(rrq, rr);
  {
    _pl_rr_destroy(rr);
  }
  pl_rr_q_LOOP_END();

  assert(rrq->count == 0);

  return 0;
}


int _pl_rr_q_complete(pl_rr_q *rrq, pl_rr *rr)
     /*@ 
       Called by endpoint when a %rr% has made a complete round trip.
     */
{
  int result = 0;

  if ( PL_RR_Q_DEBUG ) {
    fprintf(stderr, "_pl_rr_q_complete(%p, %p)\n", rrq, rr);
  }

  assert(rrq);
  assert(rr);

  /* If both buffers are complete and not multipart. */
  if ( (rr->buf[rr->dir].state == pl_rr_buf_IO_COMPLETE) &&
       (rr->buf[1 - rr->dir].state = pl_rr_buf_COMPLETE) &&
       rr->buf[0].part == 0 && 
       rr->buf[1].part == 0 ) {

    if ( PL_RR_Q_DEBUG >= 2 ) {
      fprintf(stderr, "DESTROY %p\n", rr);
      pl_rr_dump(rr, stderr);
    }
  
    /* Can be destroyed. */
    _pl_rr_destroy(rr);
  }

  return result;
}



static int _pl_rr_q_check_count(pl_rr_q *rrq)
     /*@ Fires 'empty' event if %rrq% is empty. */
{
  /* If it's already empty do it now. */
  if ( rrq->count == 0 ) {
    if ( rrq->event_empty ) {
      rrq->event_empty(rrq);
      
      /* Do it only once. */
      rrq->event_empty = 0;
    }
  }

  return 0;
}


int pl_rr_q_set_event_empty(pl_rr_q *rrq, pl_rr_q_event_handler eh)
     /*@
       Set empty event handler.
       Do not rrq->event_empty = foo.
     */
{
  assert(rrq);

  rrq->event_empty = eh;

  _pl_rr_q_check_count(rrq);

  return 0;
}


int _pl_rr_q_remove(pl_rr_q *rrq, pl_rr *rr)
     /*@ Removes %rr% from %rrq%.  May cause empty event. */
       
{
  assert(rrq);
  assert(rr);

  if ( rr->queue ) {    
    assert(rr->queue == rrq);
    
    if ( PL_RR_Q_DEBUG >= 3 ) {
      fprintf(stderr, "_pl_rr_q_remove(%p, %p)\n", rrq, rr);
    }

    pl_dll_REMOVE(rr);
    
    rr->queue->count --;
    assert(rr->queue->count >= 0);
    rr->queue = 0;

    /* Notify if queue empty. */
    _pl_rr_q_check_count(rrq);
  }

  return 0;
}


int _pl_rr_q_add_to_end(pl_rr_q *rrq, pl_rr *rr)
     /*@ 
       Adds %rr% to end of %rrq%. 
     */
{
  pl_rr *end;

  assert(rrq);
  assert(rr);

  if ( PL_RR_Q_DEBUG >= 3 ) {    
    fprintf(stderr, "_pl_rr_q_add_to_end(%p, %p):\n", rrq, rr);
    pl_rr_q_dump(rrq, stderr);
  }

  ++ rrq->count; /* Prevent rrq->queue_empty(rrq) from getting called. */

  _pl_rr_q_remove(rrq, rr);

  end = pl_rr_q_end(rrq);

  pl_dll_INSERT_BEFORE(rr, end);

  rr->queue = rrq;
  assert(rr->queue->count >= 0);
  ++ rr->queue->count;

  -- rrq->count; /* Prevent rrq->queue_empty(rrq) from getting called. */

  if ( PL_RR_Q_DEBUG >= 3 ) {    
    fprintf(stderr, "_pl_rr_q_add_to_end(%p, %p): DONE\n", rrq, rr);
    pl_rr_q_dump(rrq, stderr);
  }

  return 0;
}


int _pl_rr_q_add_to_begin(pl_rr_q *rrq, pl_rr *rr)
     /*@
       Adds %rr% to beginning of %rrq%.
      */
{
  pl_rr *beg;

  assert(rrq);
  assert(rr);
 
  if ( PL_RR_Q_DEBUG >= 3 ) {    
    fprintf(stderr, "_pl_rr_q_add_to_begin(%p, %p):\n", rrq, rr);
    pl_rr_q_dump(rrq, stderr);
  }

  ++ rrq->count; /* Prevent rrq->queue_empty(rrq) from getting called. */

  _pl_rr_q_remove(rrq, rr);

  beg = pl_rr_q_begin(rrq);

  pl_dll_INSERT_BEFORE(rr, beg);

  rr->queue = rrq;
  assert(rr->queue->count >= 0);
  ++ rr->queue->count;

  -- rrq->count; /* Prevent rrq->queue_empty(rrq) from getting called. */

  if ( PL_RR_Q_DEBUG >= 3 ) {    
    fprintf(stderr, "_pl_rr_q_add_to_begin(%p, %p): DONE\n", rrq, rr);
    pl_rr_q_dump(rrq, stderr);
  }

  return 0;
}


pl_rr *pl_rr_q_req_open(pl_rr_q *rrq, pl_rr_endpoint *dst, pl_rr_handler resh, void *data)
     /*@
       Creates a new outgoing request to %destination%.
       %resh% will be called with %data% when a response for the request has been received.
       If %resh% is 0, the request is sent, but the response is ignored.
       Returns a new request object.
     */
{
  pl_rr *rr;

  assert(rrq);

  rr = malloc(sizeof(*rr));
  assert(rr);

  pl_rr_init(rr);

  rr->dir = pl_rr_OUTGOING;
  rr->ep[rr->dir] = dst;
  rr->id = ++ rrq->head.id;
  rr->resh = resh;
  rr->resh_data = data;

  _pl_rr_q_add_to_end(rrq, rr);

  return rr;
}


pl_rr *_pl_rr_q_begin_read(pl_rr_q *rrq, int op, pl_rr_id id, pl_rr_id part, size_t len, pl_rr_endpoint *ep)
     /*@ Locate or create an rr for reading. */
{
  pl_rr *rr;

  assert(rrq);

  switch ( op ) {
  case pl_rr_REQUEST:
    /* Allocate a new request. */
    rr = malloc(sizeof(*rr));
    assert(rr);
    
    pl_rr_init(rr);
    
    rr->id = id;
    rr->dir = pl_rr_INCOMING;
    rr->ep[rr->dir] = ep;

    /* Resize request buffer. */
    pl_rr_buf_resize(&(rr->buf[op]), len); 
    
    _pl_rr_q_add_to_end(rrq, rr);
    break;

  case pl_rr_RESPONSE:
    /* src is responding to a previously sent request. */
    /* Look for a matching OUTBOUND request object that has been sent to src. */
    for ( rr = pl_rr_q_begin(rrq); rr != pl_rr_q_end(rrq); rr = rr->next ) {
      if ( rr->id == id && 
	   rr->dir == pl_rr_OUTGOING && 
	   rr->ep[rr->dir] == ep
	   ) {
	break;
      }
    }
    
    if ( rr == pl_rr_q_end(rrq) ) {
      /* Create a dummy request to read the request.  */
      fprintf(stderr, "RESPONSE %lu from %p without REQUEST\n", id, ep);

      rr = malloc(sizeof(*rr));
      pl_rr_init(rr);

      rr->id = id;
      rr->dir = pl_rr_OUTGOING; 
      rr->ep[rr->dir] = ep;
      rr->buf[pl_rr_REQUEST].state = pl_rr_buf_IO_COMPLETE;

      _pl_rr_q_add_to_end(rrq, rr);
    }

    /* Prepare response read buffer. */
    {
      pl_rr_buf *rrb = &(rr->buf[op]);
      
      rrb->part = part; 
      pl_rr_buf_resize(rrb, len);
      _pl_rr_buf_reset(rrb);
    }
    break;

  default:
    abort();
    break;
  }

  return rr;
}


/*************************************************************************/


pl_rr *_pl_rr_q_next_sendable(pl_rr_q *rrq, pl_rr_endpoint *ep)
     /* Locates next sendable rr for %ep%. */
{
  pl_rr *rr;

  for ( rr = pl_rr_q_begin(rrq); rr != pl_rr_q_end(rrq); rr = rr->next ) {
    assert(0 <= rr->dir && rr->dir <= 1);    
    if ( rr->ep[rr->dir] == ep && 
	 pl_rr_buf_get_state(&rr->buf[rr->dir]) <= pl_rr_buf_IO_PENDING ) {
      if ( PL_RR_Q_DEBUG ) {
	fprintf(stderr, "_pl_rr_q_next_sendable(%p, %p) = %p\n", rrq, ep, rr);
	pl_rr_q_dump(rrq, stderr);
      }

      return rr;
    }
  }

  return 0;
}


/*************************************************************************/


int _pl_rr_q_reschedule(pl_rr_q *rrq, pl_rr_endpoint *ep)
     /*@ Reschedule any rrs for %ep% that will need resending. */
{
  pl_rr *rr;

  assert(rrq);
  assert(ep);

  pl_rr_q_LOOP(rrq, rr);
  {
    pl_rr_buf *rrb;

    rr_next = rr->next;

    switch ( rr->dir ) {
    case pl_rr_OUTGOING:
      /*
      ** If outgoing request was sent and response has not been received,
      **  reset rr for request resend.
      ** Rational:
      **  This is a client and the server lost its connection;
      **  the client will resend.
      */
      rrb = &rr->buf[pl_rr_REQUEST];
      if ( rr->ep[rr->dir] == ep && 
	   rrb->state <= pl_rr_buf_IO_COMPLETE && 
	   rr->buf[pl_rr_RESPONSE].state <= pl_rr_buf_IO_PENDING ) {
	rrb->readp = rrb->buf;
	rrb->state = pl_rr_buf_COMPLETE;
	
	_pl_rr_buf_reset(&rr->buf[pl_rr_RESPONSE]);
      }
      break;
      
    case pl_rr_INCOMING:
      /*
      ** If incoming request was received and read (COMPLETE) and response has not been sent,
      **   assume requestor will resend.
      **
      ** Rational:
      **  This is a server and the client lost its connection;
      *   the client will resend.
      */
      _pl_rr_destroy(rr);
      break;
    }
  }
  pl_rr_q_LOOP_END();

  return 0;
}



/******************************************************************/

int pl_rr_q_dump(pl_rr_q *rrq, FILE *fp)
     /*@ Dump an %rrq% to %fp%. */
{
  pl_rr *rr;

  assert(rrq);
  assert(fp);

  fprintf(fp, "pl_rr_q %p {\n", (void*) rrq);

  pl_rr_q_LOOP(rrq, rr);
  {
    pl_rr_dump(rr, fp);
  }
  pl_rr_q_LOOP_END();

  fprintf(fp, "}\n");

  return 0;
}


/******************************************************************/

#ifdef pl_rr_q_begin
#undef pl_rr_q_begin
#endif

pl_rr *pl_rr_q_begin(pl_rr_q *rrq)
     /*@
       Returns the first pl_rr object in the %rrq%.
     */
{
  assert(rrq);

  return pl_dll_BEGIN(&(rrq->head));
}


#ifdef pl_rr_q_end
#undef pl_rr_q_end
#endif
pl_rr *pl_rr_q_end(pl_rr_q *rrq)
     /*@
       Returns the end of the %rrq%; this is not a valid pl_rr object.
     */
{
  assert(rrq);

  return pl_dll_END(&(rrq->head));
}


/******************************************************************/
