/* Author: kurt.stephens@parlano.com 2000/09/25 */
/*   parlano eventloop library -- pl_rr/reqres.c
 *   Copyright (C) 2001 Parlano, Inc.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "reqres.h"

#include "pl_el/dlinklist.h"
#include "pl_el/pl_el.h" /* pl_delayed_free() */

#include <stdlib.h> /* malloc(), free(), realloc() */
#include <stdio.h>
#include <assert.h>


/*************************************************************************/


int pl_rr_init(pl_rr *rr)
     /*@ Initialize %rr%. */
{
  assert(rr);

  memset(rr, 0, sizeof(*rr));

  pl_dll_INIT(rr);

  pl_rr_buf_init(&(rr->buf[pl_rr_REQUEST]), 0, 0, 0);
  pl_rr_buf_init(&(rr->buf[pl_rr_RESPONSE]), 0, 0, 0);

  return 0;
}


int pl_rr_req_init(pl_rr *rr, pl_rr_id id, pl_rr_id part, size_t len)
     /*@ Initialize %rr% for request %id% of length %len%. */
{
  assert(rr);

  rr->id = id;
  pl_rr_buf_init(&(rr->buf[pl_rr_REQUEST]), part, 0, len);

  return 0;
}


int _pl_rr_destroy(pl_rr *rr)
     /*@ Destroys contents of %rr%. */
{
  assert(rr);

  assert(rr->state != pl_rr_DESTROYED);

  /*
    If the request is not in a closed state
    it must have been aborted.
  */
  if ( rr->state < pl_rr_CLOSED ) {
    rr->state = pl_rr_ABORTED;
  }

  /* Event handler. */
  if ( rr->event_destroy ) {
    /* Call only once. */
    pl_rr_event_handler x = rr->event_destroy;
    rr->event_destroy = 0;
    x(rr);
  }

  /* Remove from queue. */
  if ( rr->queue ) {
    _pl_rr_q_remove(rr->queue, rr);
  }

  /* Destroy buffers. */
  pl_rr_buf_destroy(&(rr->buf[pl_rr_REQUEST]));
  pl_rr_buf_destroy(&(rr->buf[pl_rr_RESPONSE]));

  /* Put it in destroyed state. */
  rr->state = pl_rr_DESTROYED;

  /* Free it later. */
  pl_delayed_free(rr);

  return 0;
}


/***************************************************************************/

int _pl_rr_close(pl_rr *rr, int op)
{
  int result = -1;
  pl_rr_buf *rrb;

  assert(rr);
  assert(rr->state > pl_rr_DESTROYED);

#if 0
  fprintf(stderr, "_pl_rr_close(%p, %d):\n", rr, op);
  pl_rr_dump(rr, stderr);
#endif

  rrb = &(rr->buf[op]);
  
  do {
    if ( (result = pl_rr_buf_close(rrb)) )
      break;

    switch ( rr->dir ) {
    case pl_rr_OUTGOING:
      switch ( op ) {
      case pl_rr_REQUEST:
	switch ( rrb->state ) {
	case pl_rr_buf_COMPLETE:
	  /* outgoing request complete: the request is sendable. */
	  result = _pl_rr_endpoint_sendable(rr->ep[rr->dir], rr);
	  break;

	case pl_rr_buf_IO_COMPLETE:
	  /* outgoing request sent: waiting for response. */
	  rr->state = pl_rr_REQUEST_IO_COMPLETE;
	  result = 0;
	  break;

	default:
	  assert(! "COMPLETE | IO_COMPLETE");
	  abort();
	  break;
	}
	break;

      case pl_rr_RESPONSE:
	/* outgoing request: response complete: */
	switch ( rrb->state ) {
	case pl_rr_buf_IO_COMPLETE:
	  /* response read complete: process response. */
	  rr->state = pl_rr_RESPONSE_COMPLETE;
	  rrb->state = pl_rr_buf_START;

	  if ( rr->resh ) {
	    /* Response wanted by callback. */
	    result = rr->resh(rr, rr->resh_data, rr->ep[rr->dir]);
	  } else {
	    /* Response not wanted by callback. */
	    pl_rr_res_close(rr);
	  }
	  break;

	case pl_rr_buf_COMPLETE:
	  /* response used by callback: terminated req/res. */
	  rr->state = pl_rr_CLOSED;
	  result = _pl_rr_q_complete(rr->queue, rr);
	  break;

	default:
	  assert(! "IO_COMPLETE | COMPLETE");
	  abort();
	  break;
	  }
	break;
	
      default:
	assert(! "REQUEST | RESPONSE");
	abort();
	break;
      }
      break;

    case pl_rr_INCOMING:
      switch ( op ) {
      case pl_rr_REQUEST:
	/* incoming request: */
	switch ( rrb->state ) {
	case pl_rr_buf_IO_COMPLETE:
	  /* incoming request: read: process request. */
	  rr->state = pl_rr_REQUEST_COMPLETE;
	  rrb->state = pl_rr_buf_START;

	  if ( rr->ep[rr->dir]->reqh ) {
	    result = rr->ep[rr->dir]->reqh(rr, rr->ep[rr->dir]->reqh_data, rr->ep[rr->dir]);
	  } else {
	    /* Unwanted request: return null. */
	    pl_rr_req_close(rr);
	    pl_rr_res_close(rr);
	    result = 0;
	  }
	  break;
	
	case pl_rr_buf_COMPLETE:
	  /* incoming request: request used. */
	  rr->state = pl_rr_CLOSED;
	  result = _pl_rr_q_complete(rr->queue, rr);
	  break;

	default:
	  assert(! "IO_COMPLETE | COMPLETE");
	  abort();
	  break;
	}
	break;

      case pl_rr_RESPONSE:
	switch ( rrb->state ) {
	case pl_rr_buf_COMPLETE:
	  /* incoming request: response complete and sendable. */
	  result = _pl_rr_endpoint_sendable(rr->ep[rr->dir], rr);
	  break;

	case pl_rr_buf_IO_COMPLETE:
	  /* response was sent: request is complete. */
	  rr->state = pl_rr_CLOSED;
	  result = _pl_rr_q_complete(rr->queue, rr);
	  break;

	default:
	  assert(! "COMPLETE | IO_COMPLETE");
	  abort();
	  break;
	}
	break;
      
      default:
	assert(! "REQUEST | RESPONSE");
	abort();
	break;
      }
      break;

    default:
      assert(! "INCOMING | OUTGOING");
      abort();
      break;
    }
  } while ( 0 );

  return result;
}


/*************************************************************************/


#ifdef pl_rr_get_state
#undef pl_rr_get_state
#endif
pl_rr_state pl_rr_get_state(pl_rr *rr)
     /*@ Returns the state of the request %rr%. */
{
  return rr->state;
}


/*************************************************************************/


size_t pl_rr_req_len(pl_rr *rr)
     /*@ Returns the length of the request %rr%. */
{
  assert(rr);
  assert(rr->state > pl_rr_DESTROYED);

  return pl_rr_buf_read_len(&(rr->buf[pl_rr_REQUEST]));
}


const char *pl_rr_req_buf(pl_rr *rr)
     /*@ Returns beginning of request buffer. */
{
  assert(rr);
  assert(rr->state > pl_rr_DESTROYED);

  return rr->buf[pl_rr_REQUEST].buf;
}


pl_rr_id  pl_rr_req_part(pl_rr *rr)
     /*@ Returns part id of request %rr%. */
{
  assert(rr);
  assert(rr->state > pl_rr_DESTROYED);

  return rr->buf[pl_rr_REQUEST].part;
}


int pl_rr_req_write(pl_rr *rr, const void *buf, size_t len)
     /*@ Appends %buf% to the request %rr%. */
{
  assert(rr);
  assert(rr->state > pl_rr_DESTROYED);

  return pl_rr_buf_write(&(rr->buf[pl_rr_REQUEST]), buf, len);
}


int pl_rr_req_close(pl_rr *rr)
     /*@ Commits request %rr% for sending. */
{
  return _pl_rr_close(rr, pl_rr_REQUEST);
}


/*************************************************************************/


size_t pl_rr_res_len(pl_rr *rr)
     /*@ Returns length of response %rr%. */
{
  assert(rr);
  assert(rr->state > pl_rr_DESTROYED);

  return pl_rr_buf_read_len(&(rr->buf[pl_rr_RESPONSE]));
}


const char *pl_rr_res_buf(pl_rr *rr)
     /*@ Returns beginning of response buffer. */
{
  assert(rr);
  assert(rr->state > pl_rr_DESTROYED);

  return rr->buf[pl_rr_RESPONSE].buf;
}


pl_rr_id  pl_rr_res_part(pl_rr *rr)
     /*@ Returns part id of response %rr%. */
{
  assert(rr);
  assert(rr->state > pl_rr_DESTROYED);

  return rr->buf[pl_rr_RESPONSE].part;
}


int pl_rr_res_write(pl_rr *rr, const void *buf, size_t len)
     /*@ Writes %buf% to response %rr%. */
{
  assert(rr);
  assert(rr->state > pl_rr_DESTROYED);

  return pl_rr_buf_write(&(rr->buf[pl_rr_RESPONSE]), buf, len);
}


int pl_rr_res_read(pl_rr *rr, void *buf, size_t len)
     /*@ 
       Reads %len% bytes from response %rr% into %buf%. 
       Returns number of bytes read.
      */
{
  size_t avail;
  
  assert(rr);
  assert(rr->state > pl_rr_DESTROYED);

  avail = pl_rr_res_len(rr);
  len = avail < len ? avail : len;
  if ( len ) {
    assert(buf);
    memcpy(buf, pl_rr_buf_read_ptr(&(rr->buf[pl_rr_RESPONSE])), len);
    pl_rr_buf_read_count(&(rr->buf[pl_rr_RESPONSE]), len);
  }

  return len;
}


int pl_rr_res_close(pl_rr *rr)
     /*@
       Commits response %rr% for transmission back to requestor.
     */
{
  return _pl_rr_close(rr, pl_rr_RESPONSE);
}


/***********************************************************/


pl_rr_endpoint *pl_rr_dst(pl_rr *rr)
     /*@
       The destination endpoint of the request.
     */
{
  assert(rr);
  assert(rr->state > pl_rr_DESTROYED);

  return rr->ep[pl_rr_OUTGOING];
}


pl_rr_endpoint *pl_rr_src(pl_rr *rr)
     /*@
       The source endpoint of the request.
     */
{
  assert(rr);
  assert(rr->state > pl_rr_DESTROYED);

  return rr->ep[pl_rr_INCOMING];
}


/***********************************************************/


int pl_rr_dump(pl_rr *rr, FILE *fp)
{
  assert(rr);
  assert(fp);

  fprintf(fp, "  rr %lu %d (%p %s %p) %p\n", 
	  rr->id,
	  rr->state,
	  rr->ep[0],
	  rr->dir == pl_rr_INCOMING ? "<=" : "=>", 
	  rr->ep[1],
	  rr);

  pl_rr_buf_dump(&rr->buf[0], "    req ", fp);
  pl_rr_buf_dump(&rr->buf[1], "    res ", fp);

  return 0;
}


/******************************************************************/
